Package: multipleOutcomes
Title: Joint Covariance and Treatment-Effect Tests for Multiple
        Outcomes
Version: 0.16.2
Authors@R: 
    person("Han", "Zhang", , "zhangh.ustc@gmail.com", role = c("aut", "cre"))
Description: Fits generalized linear models, Cox proportional-hazards
    models, log-rank tests, generalized estimating equations, mixed models
    with repeated measures, Kaplan-Meier curves, and quantile differences
    jointly across multiple endpoints, and returns the full asymptotic
    covariance matrix linking them. Implements PATED (Prognostic Assisted 
    Treatment Effect Detection), a randomized-trial method that
    exploits balanced prognostic covariates to tighten standard errors and
    increase statistical power without introducing bias.
URL: https://github.com/zhangh12/multipleOutcomes,
        https://zhangh12.github.io/multipleOutcomes/
BugReports: https://github.com/zhangh12/multipleOutcomes/issues
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: dplyr, ggplot2, ggpubr, mmrm (>= 0.3.15), mvtnorm, rlang,
        sandwich, stringr, survival, tidyr, tidyselect
Suggests: asaur, coin, iBST, invGauss, JM, joint.Cox, knitr, momentfit,
        numDeriv, pec, randomForestSRC, rmarkdown, testthat (>= 3.0.0)
Depends: R (>= 3.5)
LazyData: true
NeedsCompilation: yes
Packaged: 2026-05-17 14:24:30 UTC; zhhan
Author: Han Zhang [aut, cre]
Maintainer: Han Zhang <zhangh.ustc@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-18 07:00:09 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-18 09:53:18 UTC; unix
Archs: multipleOutcomes.so.dSYM

Package: multipleOutcomes
Title: Joint Covariance and Treatment-Effect Tests for Multiple
        Outcomes
Version: 0.16.2
Authors@R: 
    person("Han", "Zhang", , "zhangh.ustc@gmail.com", role = c("aut", "cre"))
Description: Fits generalized linear models, Cox proportional-hazards
    models, log-rank tests, generalized estimating equations, mixed models
    with repeated measures, Kaplan-Meier curves, and quantile differences
    jointly across multiple endpoints, and returns the full asymptotic
    covariance matrix linking them. Implements PATED (Prognostic Assisted 
    Treatment Effect Detection), a randomized-trial method that
    exploits balanced prognostic covariates to tighten standard errors and
    increase statistical power without introducing bias.
URL: https://github.com/zhangh12/multipleOutcomes,
        https://zhangh12.github.io/multipleOutcomes/
BugReports: https://github.com/zhangh12/multipleOutcomes/issues
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: dplyr, ggplot2, ggpubr, mmrm (>= 0.3.15), mvtnorm, rlang,
        sandwich, stringr, survival, tidyr, tidyselect
Suggests: asaur, coin, iBST, invGauss, JM, joint.Cox, knitr, momentfit,
        numDeriv, pec, randomForestSRC, rmarkdown, testthat (>= 3.0.0)
Depends: R (>= 3.5)
LazyData: true
NeedsCompilation: yes
Packaged: 2026-05-17 14:24:30 UTC; zhhan
Author: Han Zhang [aut, cre]
Maintainer: Han Zhang <zhangh.ustc@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-18 07:00:09 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-18 10:06:54 UTC; unix
Archs: multipleOutcomes.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(coef,jointCovariance)
S3method(jointCovariance,default)
S3method(plot,pated)
S3method(print,summary.jointCovariance)
S3method(summary,jointCovariance)
S3method(vcov,jointCovariance)
export(coxph_)
export(gee_)
export(glm_)
export(jointCovariance)
export(km_)
export(logrank_)
export(mmrm_)
export(pated)
export(quantile_)
export(simulateMoData)
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_when)
importFrom(dplyr,desc)
importFrom(dplyr,filter)
importFrom(dplyr,first)
importFrom(dplyr,group_by)
importFrom(dplyr,mutate)
importFrom(dplyr,row_number)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(ggplot2,aes)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_ribbon)
importFrom(ggplot2,geom_step)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,ylim)
importFrom(ggpubr,ggarrange)
importFrom(mvtnorm,rmvnorm)
importFrom(rlang,.data)
importFrom(sandwich,bread)
importFrom(sandwich,estfun)
importFrom(stats,coef)
importFrom(stats,cov)
importFrom(stats,cov2cor)
importFrom(stats,gaussian)
importFrom(stats,model.extract)
importFrom(stats,model.matrix)
importFrom(stats,pchisq)
importFrom(stats,printCoefmat)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,resid)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,time)
importFrom(stats,var)
importFrom(stats,vcov)
importFrom(stringr,str_extract)
importFrom(stringr,str_glue)
importFrom(survival,strata)
importFrom(survival,survfit)
importFrom(tidyr,pivot_longer)
importFrom(tidyselect,all_of)
importFrom(utils,capture.output)
useDynLib(multipleOutcomes, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# multipleOutcomes

<!-- badges: start -->

[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![Code
Coverage](https://codecov.io/gh/zhangh12/multipleOutcomes/branch/main/graph/badge.svg)](https://app.codecov.io/gh/zhangh12/multipleOutcomes)
<!-- badges: end -->

The goal of multipleOutcomes is to fit statistical models for multiple
outcomes simultaneously. It computes estimates of parameters across
fitted models and returns the matrix of asymptotic covariance. Various
applications of this package, including PATED (Prognostic Variables
Assisted Treatment Effect Detection), multiple comparison adjustment,
conditional power are illustrated.

## Installation

Install the development version from GitHub:

``` r
if(!requireNamespace("remotes", quietly = TRUE)) {
  install.packages("remotes")
}
remotes::install_github(
  "zhangh12/multipleOutcomes",
  dependencies = TRUE
)
```

## Example

Let’s get started by analyzing a real randomized trial data. The data
was from a randomized clinical trial of indomethacin to prevent
post-ERCP pancreatitis. The binary endpoint is post-ERCP pancreatitis
(yes/no). It was collected, reformatted and shared by Dr. Peter Higgins
in his R package
[`medicaldata`](https://CRAN.R-project.org/package=medicaldata). For
more information, please refer to the data set `indo_rct` in
`medicaldata`, and the
[manuscript](https://doi.org/10.1056/NEJMoa1111103).

Here we use the PATED method to estimate marginal treatment effect on
post-ERCP pancreatitis, adjusting for a set of risk factors for PEP as
potential prognostic covariates. I am not an expert of this disease, so
I simply follow the description information of the covariates in
`medicaldata`.

Here we compare PATED with unconditional logistic regression model in
estimating the treatment effect. For each of the covariates, a
regression model `covar ~ rx` is fitted to compute the slope coefficient
$\widehat b$. The correlation between $\widehat b$ and $\widehat\beta$,
the unadjusted estimate of treatment effect (`outcome ~ rx`), are
computed in the column `corr`.

``` r
library(multipleOutcomes)
library(dplyr)

options(digits = 4)

data(indo)

# jointCovariance() and pated() need a per-row subject identifier in
# column `pid`. The trial data already has an integer `id` — we just
# stamp it as a string under the conventional name.
indo <- indo %>% mutate(pid = paste0("s-", id))

# When there's only one dataset, `data_index` defaults to 1 and the
# `list()` wrap is auto-applied, so neither needs to be spelled out.
fit <- pated(
  glm_(outcome     ~ rx, family = "binomial"),

  glm_(risk        ~ rx, family = "gaussian"),
  glm_(gender      ~ rx, family = "binomial"),
  glm_(sod         ~ rx, family = "binomial"),
  glm_(pep         ~ rx, family = "binomial"),
  glm_(recpanc     ~ rx, family = "binomial"),
  glm_(psphinc     ~ rx, family = "binomial"),
  glm_(precut      ~ rx, family = "binomial"),
  glm_(difcan      ~ rx, family = "binomial"),
  glm_(amp         ~ rx, family = "binomial"),
  glm_(paninj      ~ rx, family = "binomial"),
  glm_(acinar      ~ rx, family = "binomial"),
  glm_(asa81       ~ rx, family = "binomial"),
  glm_(asa         ~ rx, family = "binomial"),
  glm_(prophystent ~ rx, family = "binomial"),
  glm_(therastent  ~ rx, family = "binomial"),
  glm_(pdstent     ~ rx, family = "binomial"),

  data = indo
)

plot(fit)
```

<img src="man/figures/README-unnamed-chunk-2-1.png" alt="" width="100%" />

The plot above shows that all covariates are very well balanced between
two arms (i.e. z-values of grey points are close to zero).

``` r
print(fit)
#>           term family  estimate  stderr   pvalue     method      corr
#> 1      outcome  PATED -0.751357 0.24697 0.002348      PATED        NA
#> 2      outcome    glm -0.705130 0.25283 0.005287   Standard  1.000000
#> 3         risk    glm  0.083338 0.07168 0.244984 Prognostic  0.139678
#> 4          pep    glm -0.002142 0.22269 0.992326 Prognostic  0.126343
#> 5          amp    glm  0.041102 0.47862 0.931565 Prognostic  0.082925
#> 6       acinar    glm  0.275329 0.39618 0.487084 Prognostic  0.057366
#> 7       paninj    glm -0.360946 0.27417 0.188011 Prognostic  0.053484
#> 8   therastent    glm -0.252931 0.31591 0.423343 Prognostic -0.049425
#> 9       difcan    glm  0.105662 0.18570 0.569368 Prognostic  0.048068
#> 10     psphinc    glm  0.133453 0.16480 0.418059 Prognostic  0.037479
#> 11 prophystent    glm  0.215272 0.18961 0.256226 Prognostic  0.034128
#> 12     recpanc    glm -0.069990 0.17817 0.694454 Prognostic  0.033145
#> 13         asa    glm -0.072330 0.27877 0.795278 Prognostic -0.024236
#> 14         sod    glm  0.248237 0.21453 0.247228 Prognostic -0.010506
#> 15       asa81    glm -0.394693 0.31587 0.211467 Prognostic -0.007874
#> 16     pdstent    glm  0.181177 0.21480 0.398967 Prognostic  0.006305
#> 17      precut    glm -0.090072 0.36402 0.804570 Prognostic  0.005952
#> 18      gender    glm  0.170977 0.20058 0.393994 Prognostic  0.004055
```

It is evident that PATED produces a treatment effect estimate (-0.7514)
that is close to the one from the unconditional logistic regression
model (-0.7051), indicating a good property of unbiasness. Meanwhile, a
slightly lower SE (0.247 v.s. 0.2528) and more significant p-value
(0.0023 v.s. 0.0053) are obtained by PATED.

In comparison, we adjust the list of covariates in a logistic regression
to estimate conditional log odds ratio of treatment. We have to exclude
`asa81` and `asa` due to missing data, and `prophystent`, `therastent`
and `pdstent` due to numerical instability.

``` r
logit <- glm(
  outcome ~ rx + risk + gender + sod + pep + recpanc + psphinc + precut + difcan + amp + paninj + acinar, 
  family = 'binomial', data = indo
) |> summary()

logit
#> 
#> Call:
#> glm(formula = outcome ~ rx + risk + gender + sod + pep + recpanc + 
#>     psphinc + precut + difcan + amp + paninj + acinar, family = "binomial", 
#>     data = indo)
#> 
#> Coefficients:
#>              Estimate Std. Error z value Pr(>|z|)    
#> (Intercept)  -2.78401    0.47052   -5.92  3.3e-09 ***
#> rx           -0.80110    0.26337   -3.04   0.0024 ** 
#> risk          0.15917    0.54108    0.29   0.7686    
#> gender2_male -0.00483    0.35756   -0.01   0.9892    
#> sod1_yes      0.33084    0.72780    0.45   0.6494    
#> pep1_yes      0.98179    0.63710    1.54   0.1233    
#> recpanc1_yes  0.01492    0.38860    0.04   0.9694    
#> psphinc1_yes  0.11582    0.59986    0.19   0.8469    
#> precut1_yes   0.20988    0.75567    0.28   0.7812    
#> difcan1_yes   0.41110    0.61653    0.67   0.5049    
#> amp1_yes      1.92303    0.85627    2.25   0.0247 *  
#> paninj1_yes   0.16629    0.47729    0.35   0.7275    
#> acinar1_yes   0.98954    0.61083    1.62   0.1052    
#> ---
#> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
#> 
#> (Dispersion parameter for binomial family taken to be 1)
#> 
#>     Null deviance: 468.01  on 601  degrees of freedom
#> Residual deviance: 433.43  on 589  degrees of freedom
#> AIC: 459.4
#> 
#> Number of Fisher Scoring iterations: 5
```

The conditional treatment effect is estimated as -0.8011, with SE 0.2634
and p-value 0.0024.

In this example, it seems like using unconditional logistic regression
or PATED can underestimate treatment effect compared to regression
adjustment. However, this is not always the case in practice. We also
see examples where the regression adjustment can underestimate the
treatment effect substantially. In fact, such a comparison is invalid
because they are estimating different parameters under different
definitions of treatment effect. In terms of statistical power of
detecting treatment effect, comprehensive simulation studies show that
PATED always outperform unconditional logistic regression, and is
comparable to (if not better than) regression adjustment.

# Reproducible simulators used by both fast smoke tests and the slower
# Monte Carlo empirical-vs-theoretical covariance test.

# Continuous outcome + one prognostic continuous covariate.
sim_gaussian <- function(n = 200,
                         beta_arm = 0.30,
                         beta_z   = 0.50,
                         sigma    = 1,
                         seed     = NULL) {
  if (!is.null(seed)) set.seed(seed)
  arm <- rbinom(n, 1, 0.5)
  z   <- rnorm(n)
  y   <- beta_arm * arm + beta_z * z + rnorm(n, sd = sigma)
  data.frame(pid = paste0("p", seq_len(n)), arm = arm, z = z, y = y)
}

# Binary outcome.
sim_binomial <- function(n = 300,
                         beta_arm = 0.50,
                         beta_z   = 0.50,
                         seed     = NULL) {
  if (!is.null(seed)) set.seed(seed)
  arm <- rbinom(n, 1, 0.5)
  z   <- rnorm(n)
  p   <- plogis(beta_arm * arm + beta_z * z)
  data.frame(pid = paste0("p", seq_len(n)), arm = arm, z = z, y = rbinom(n, 1, p))
}

# Exponential survival under a proportional-hazards model with admin censoring.
sim_coxph <- function(n = 300,
                      beta_arm  = -0.40,
                      beta_z    = -0.30,
                      lam0      = 0.30,
                      admin_tau = 5,
                      seed      = NULL) {
  if (!is.null(seed)) set.seed(seed)
  arm <- rbinom(n, 1, 0.5)
  z   <- rnorm(n)
  lam <- lam0 * exp(beta_arm * arm + beta_z * z)
  t   <- rexp(n, rate = lam)
  event <- as.integer(t <= admin_tau)
  time  <- pmin(t, admin_tau)
  data.frame(pid = paste0("p", seq_len(n)), arm = arm, z = z,
             time = time, event = event)
}

# Multi-type correlated outcomes. A single subject contributes:
#   - a continuous outcome y_cont
#   - a binary outcome y_bin
#   - a right-censored survival outcome (time, event)
#   - a longitudinal trajectory y_long with `n_visits` rows
# Correlation across endpoints is induced by a shared latent vector
# u ~ N(0, Sigma) with Sigma having exchangeable correlation `rho`. The
# treatment `arm` is randomized independently. Returns a list with
# `wide` (one row per subject) and `long` (one row per subject-visit),
# both keyed on the same `pid`. With rho = 0 the endpoints are
# independent given `arm`; with rho > 0 the joint covariance of the
# four `arm`-coefficients has non-trivial off-diagonal blocks, which is
# what jointCovariance() is supposed to estimate.
sim_correlated_multi <- function(n_subj    = 250,
                                 n_visits  = 3,
                                 beta_cont = 0.30,
                                 beta_bin  = 0.50,
                                 beta_tte  = -0.40,
                                 beta_long = 0.30,
                                 alpha     = 0.80,
                                 rho       = 0.50,
                                 lam0      = 0.30,
                                 admin_tau = 5,
                                 sigma_y   = 0.50,
                                 sigma_e   = 0.30,
                                 long_overlap_frac = 1.0,
                                 seed      = NULL) {
  if (!is.null(seed)) set.seed(seed)

  pid <- paste0("p", seq_len(n_subj))
  arm <- rbinom(n_subj, 1, 0.5)

  Sigma <- matrix(rho, 4, 4); diag(Sigma) <- 1
  u <- mvtnorm::rmvnorm(n_subj, sigma = Sigma)

  y_cont <- beta_cont * arm + alpha * u[, 1] + rnorm(n_subj, sd = sigma_y)
  y_bin  <- rbinom(n_subj, 1, plogis(beta_bin * arm + alpha * u[, 2]))
  lam    <- lam0 * exp(beta_tte * arm + alpha * u[, 3])
  t_evt  <- rexp(n_subj, rate = lam)
  time   <- pmin(t_evt, admin_tau)
  event  <- as.integer(t_evt <= admin_tau)

  wide <- data.frame(pid = pid, arm = arm,
                     y_cont = y_cont, y_bin = y_bin,
                     time = time, event = event)

  b <- alpha * u[, 4]
  visit_offset <- 0.10 * (seq_len(n_visits) - 1)
  long_y <- outer(beta_long * arm + b, rep(1, n_visits)) +
            matrix(visit_offset, n_subj, n_visits, byrow = TRUE) +
            matrix(rnorm(n_subj * n_visits, sd = sigma_e), n_subj, n_visits)

  long <- data.frame(
    pid   = rep(pid, each = n_visits),
    visit = factor(rep(seq_len(n_visits), times = n_subj)),
    arm   = rep(arm, each = n_visits),
    y_long = as.vector(t(long_y))
  )

  # Optional: keep longitudinal records for only a random subset of subjects,
  # so wide and long share only `long_overlap_frac * n_subj` pids in common.
  # Exercises the n_shared_sample_sizes / FisherInformation overlap-weighting
  # path in jointCovariance().
  if (long_overlap_frac < 1.0) {
    n_keep   <- max(1L, floor(long_overlap_frac * n_subj))
    keep_pid <- sample(pid, size = n_keep)
    long <- long[long$pid %in% keep_pid, , drop = FALSE]
  }

  list(wide = wide, long = long)
}

# Repeated measurements with within-subject AR(1) correlation.
# Returns one row per (subject, visit) with `pid` as the cluster id.
sim_repeated <- function(n_subj = 120,
                         n_visits = 3,
                         beta_arm = 0.30,
                         beta_visit = 0.10,
                         beta_z = 0.50,
                         rho = 0.5,
                         sigma = 1,
                         seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  arm <- rbinom(n_subj, 1, 0.5)
  # `z` is a per-subject prognostic baseline, sampled independently of arm.
  # Including it lets the longitudinal smoke tests pair the longitudinal
  # primary model with an arm-independent glm_(z ~ arm) prognostic — the
  # assumption PATED actually relies on for valid inference.
  z <- rnorm(n_subj)
  Sigma <- sigma^2 * rho^abs(outer(seq_len(n_visits), seq_len(n_visits), "-"))
  e <- mvtnorm::rmvnorm(n_subj, sigma = Sigma)
  visit_offset <- beta_visit * (seq_len(n_visits) - 1)
  y <- outer(beta_arm * arm + beta_z * z, rep(1, n_visits)) +
       matrix(visit_offset, n_subj, n_visits, byrow = TRUE) + e
  long <- data.frame(
    pid   = rep(paste0("p", seq_len(n_subj)), each = n_visits),
    visit = factor(rep(seq_len(n_visits), times = n_subj)),
    arm   = rep(arm, each = n_visits),
    z     = rep(z, each = n_visits),
    y     = as.vector(t(y))
  )
  long
}

# Monte Carlo validation: for several model types, simulate K independent
# datasets, fit jointCovariance() on each, and check that the empirical
# covariance of the coefficient vector across replicates agrees with the
# theoretical covariance that jointCovariance() reports.
#
# These tests are slow (hundreds of fits each) and statistical (no exact
# answer). They run under devtools::test() / R CMD check --as-cran with the
# right flags, but are skipped on CRAN and in the default test run.

# mc_enabled(), mc_run(), and expect_cov_close() live in helper-mc.R.

# ---------------------------------------------------------------------------
test_that("MC: empirical cov matches theoretical for glm_(gaussian) joint", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  K <- 400; n <- 250

  res <- mc_run(
    K,
    simulate = function(k) sim_gaussian(n = n, seed = 1000L + k),
    fit_fn   = function(dat) jointCovariance(
      glm_(y ~ arm + z, family = "gaussian", data_index = 1),
      data = list(dat)
    )
  )
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.20, abs_offdiag = 0.02,
                   label = "glm gaussian joint")
})

# ---------------------------------------------------------------------------
test_that("MC: glm_(gaussian) + glm_(gaussian) joint matrix is calibrated", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  K <- 400; n <- 250

  res <- mc_run(
    K,
    simulate = function(k) sim_gaussian(n = n, seed = 2000L + k),
    fit_fn   = function(dat) jointCovariance(
      glm_(y ~ arm, family = "gaussian", data_index = 1),
      glm_(z ~ arm, family = "gaussian", data_index = 1),
      data = list(dat)
    )
  )
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.25, abs_offdiag = 0.03,
                   label = "two-glm joint")
})

# ---------------------------------------------------------------------------
test_that("MC: empirical cov matches theoretical for glm_(binomial)", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  K <- 400; n <- 500

  res <- mc_run(
    K,
    simulate = function(k) sim_binomial(n = n, seed = 3000L + k),
    fit_fn   = function(dat) jointCovariance(
      glm_(y ~ arm + z, family = "binomial", data_index = 1),
      data = list(dat)
    )
  )
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.25, abs_offdiag = 0.03,
                   label = "glm binomial")
})

# ---------------------------------------------------------------------------
test_that("MC: logrank_ is standardized (mean 0, variance 1) under H0", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  skip_if_not_installed("survival")
  library(survival)
  K <- 400; n <- 500

  # LogRankAdapter returns I^{-1/2} * U(0): standardized log-rank stat.
  # Under H0 (beta_arm = 0) this is approximately N(0, 1).
  res <- mc_run(
    K,
    simulate = function(k) sim_coxph(n = n, beta_arm = 0,
                                     seed = 4500L + k),
    fit_fn   = function(dat) jointCovariance(
      logrank_(Surv(time, event) ~ arm, data_index = 1),
      data = list(dat)
    )
  )
  emp_var <- as.numeric(cov(res$beta_hat))
  thr_var <- as.numeric(res$vcov_mean)

  expect_lt(abs(mean(res$beta_hat)), 0.10,
            label = "logrank null mean ~ 0")
  expect_lt(abs(emp_var - 1) , 0.20,
            label = "logrank null variance ~ 1 (empirical)")
  expect_lt(abs(thr_var - 1), 0.20,
            label = "logrank null variance ~ 1 (theoretical)")
  expect_lt(abs(emp_var - thr_var) / thr_var, 0.20,
            label = "logrank empirical/theoretical variance agree")
})

# ---------------------------------------------------------------------------
test_that("MC: empirical cov matches theoretical for coxph_", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  skip_if_not_installed("survival")
  library(survival)
  K <- 400; n <- 500

  res <- mc_run(
    K,
    simulate = function(k) sim_coxph(n = n, seed = 4000L + k),
    fit_fn   = function(dat) jointCovariance(
      coxph_(Surv(time, event) ~ arm + z, data_index = 1),
      data = list(dat)
    )
  )
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.25, abs_offdiag = 0.03,
                   label = "coxph")
})

# ---------------------------------------------------------------------------
test_that("MC: pated() variance reduction is honest (calibrated SE & coverage)", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  K <- 400; n <- 250

  est_pated <- numeric(K); se_pated <- numeric(K)
  est_std   <- numeric(K); se_std   <- numeric(K)

  for (k in seq_len(K)) {
    # arm coefficient is 0 by construction; z is the only prognostic signal.
    dat <- sim_gaussian(n = n, beta_arm = 0, beta_z = 0.7,
                       seed = 5000L + k)
    pa <- pated(
      glm_(y ~ arm, family = "gaussian", data_index = 1),
      glm_(z ~ arm, family = "gaussian", data_index = 1),
      data = list(dat)
    )
    p_row <- pa[pa$method == "PATED",    ]
    s_row <- pa[pa$method == "Standard", ]
    est_pated[k] <- p_row$estimate; se_pated[k] <- p_row$stderr
    est_std[k]   <- s_row$estimate; se_std[k]   <- s_row$stderr
  }

  # Both estimators are unbiased for the null effect.
  expect_lt(abs(mean(est_pated)), 0.05,
            label = "PATED mean estimate near 0")
  expect_lt(abs(mean(est_std)),   0.05,
            label = "Standard mean estimate near 0")

  # SE reported by pated() ~ empirical SD of the estimates.
  expect_lt(abs(mean(se_pated) - sd(est_pated)) / sd(est_pated),
            0.15, label = "PATED SE calibration")
  expect_lt(abs(mean(se_std)   - sd(est_std))   / sd(est_std),
            0.15, label = "Standard SE calibration")

  # 95% CI coverage close to nominal (within MC tolerance ~ 2*sqrt(.95*.05/K)).
  cov_p <- mean(abs(est_pated / se_pated) < 1.96)
  cov_s <- mean(abs(est_std   / se_std)   < 1.96)
  expect_gt(cov_p, 0.92); expect_lt(cov_p, 0.98)
  expect_gt(cov_s, 0.92); expect_lt(cov_s, 0.98)

  # The whole point of PATED: SE smaller than unadjusted when z is prognostic.
  expect_lt(mean(se_pated), mean(se_std))
})

# Monte Carlo validation of pated() for heterogeneous primary + prognostic
# pairings (cox × glm and binomial-glm × glm). The existing
# test-mc-covariance.R already covers the same-engine gaussian case; here we
# specifically exercise the cross-engine cov21 block that pated() uses for
# its prognostic adjustment, then check that the resulting SE is calibrated
# and the 95% CI coverage is nominal.
#
# Helpers (mc_enabled, mc_run, expect_cov_close) come from helper-mc.R.

# Generic engine-agnostic calibration assertion for a pated() result. Pass
# the K-vector of point estimates, the K-vector of reported SEs, plus the
# *true* null value (typically 0 for the simulators we use here).
expect_pated_calibrated <- function(est, se, true = 0,
                                    bias_tol      = 0.05,
                                    se_rel_tol    = 0.15,
                                    cov_lo        = 0.92,
                                    cov_hi        = 0.98,
                                    label         = "") {
  expect_lt(abs(mean(est) - true), bias_tol,
            label = paste("mean estimate", label))
  expect_lt(abs(mean(se) - sd(est)) / sd(est), se_rel_tol,
            label = paste("SE calibration", label))
  cover <- mean(abs((est - true) / se) < 1.96)
  expect_gt(cover, cov_lo, label = paste("CI coverage lower", label))
  expect_lt(cover, cov_hi, label = paste("CI coverage upper", label))
}

# ---------------------------------------------------------------------------
test_that("MC: pated(coxph_, glm_) is unbiased + calibrated under H0", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  skip_if_not_installed("survival")
  library(survival)
  K <- 400; n <- 400

  est_p <- numeric(K); se_p <- numeric(K)
  est_s <- numeric(K); se_s <- numeric(K)

  for (k in seq_len(K)) {
    dat <- sim_coxph(n = n, beta_arm = 0, beta_z = -0.5,
                     seed = 6000L + k)
    pa <- pated(
      coxph_(Surv(time, event) ~ arm,                       data_index = 1),
      glm_(z ~ arm, family = "gaussian",                    data_index = 1),
      data = list(dat)
    )
    p_row <- pa[pa$method == "PATED",    ]
    s_row <- pa[pa$method == "Standard", ]
    est_p[k] <- p_row$estimate; se_p[k] <- p_row$stderr
    est_s[k] <- s_row$estimate; se_s[k] <- s_row$stderr
  }

  expect_pated_calibrated(est_p, se_p, true = 0, label = "PATED cox×glm")
  expect_pated_calibrated(est_s, se_s, true = 0, label = "Standard cox")
  # Prognostic z should tighten the SE.
  expect_lt(mean(se_p), mean(se_s))
})

# ---------------------------------------------------------------------------
test_that("MC: pated(glm_binomial, glm_) is unbiased + calibrated under H0", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  K <- 400; n <- 500

  est_p <- numeric(K); se_p <- numeric(K)
  est_s <- numeric(K); se_s <- numeric(K)

  for (k in seq_len(K)) {
    dat <- sim_binomial(n = n, beta_arm = 0, beta_z = 0.8,
                        seed = 7000L + k)
    pa <- pated(
      glm_(y ~ arm, family = "binomial", data_index = 1),
      glm_(z ~ arm, family = "gaussian", data_index = 1),
      data = list(dat)
    )
    p_row <- pa[pa$method == "PATED",    ]
    s_row <- pa[pa$method == "Standard", ]
    est_p[k] <- p_row$estimate; se_p[k] <- p_row$stderr
    est_s[k] <- s_row$estimate; se_s[k] <- s_row$stderr
  }

  expect_pated_calibrated(est_p, se_p, true = 0, label = "PATED bin×glm")
  expect_pated_calibrated(est_s, se_s, true = 0, label = "Standard binomial")
  expect_lt(mean(se_p), mean(se_s))
})

# Fast smoke tests for every supported model engine going through both
# jointCovariance() and pated(). These run in well under a second each.

# ---------------------------------------------------------------------------
# Shared assertion: a fit returned by jointCovariance() is well-formed.
expect_well_formed_jc <- function(fit, expected_models) {
  expect_s3_class(fit, "jointCovariance")
  expect_true(all(is.finite(coef(fit))))
  V <- vcov(fit)
  expect_true(is.matrix(V))
  expect_equal(dim(V), c(length(coef(fit)), length(coef(fit))))
  expect_lt(max(abs(V - t(V))), 1e-8)  # symmetric
  expect_gte(min(eigen(V, symmetric = TRUE, only.values = TRUE)$values), -1e-8)
  expect_equal(length(fit$id_map), expected_models)
}

# A pated() fit has 1 PATED row, 1 Standard row, and >=1 Prognostic row.
expect_well_formed_pated <- function(fit) {
  expect_s3_class(fit, "pated")
  expect_true(all(c("PATED", "Standard") %in% fit$method))
  expect_true(any(grepl("^Prognostic", fit$method)))
  expect_true(all(is.finite(fit$estimate)))
  expect_true(all(is.finite(fit$stderr)))
  expect_true(all(is.finite(fit$pvalue)))
}

# ---------------------------------------------------------------------------
test_that("glm_ gaussian works through jointCovariance + pated", {
  dat <- sim_gaussian(n = 200, seed = 1)
  jc <- jointCovariance(
    glm_(y ~ arm, family = "gaussian", data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_jc(jc, expected_models = 2)

  pa <- pated(
    glm_(y ~ arm, family = "gaussian", data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_pated(pa)
})

# ---------------------------------------------------------------------------
test_that("glm_ binomial works through jointCovariance + pated", {
  dat <- sim_binomial(n = 400, seed = 2)
  jc <- jointCovariance(
    glm_(y ~ arm, family = "binomial", data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_jc(jc, expected_models = 2)

  pa <- pated(
    glm_(y ~ arm, family = "binomial", data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_pated(pa)
})

# ---------------------------------------------------------------------------
test_that("coxph_ works through jointCovariance + pated", {
  skip_if_not_installed("survival")
  library(survival)
  dat <- sim_coxph(n = 400, seed = 3)
  jc <- jointCovariance(
    coxph_(Surv(time, event) ~ arm, data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_jc(jc, expected_models = 2)

  pa <- pated(
    coxph_(Surv(time, event) ~ arm, data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_pated(pa)
})

# ---------------------------------------------------------------------------
test_that("logrank_ works through jointCovariance + pated", {
  skip_if_not_installed("survival")
  library(survival)
  dat <- sim_coxph(n = 400, seed = 4)
  jc <- jointCovariance(
    logrank_(Surv(time, event) ~ arm, data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_jc(jc, expected_models = 2)

  pa <- pated(
    logrank_(Surv(time, event) ~ arm, data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat)
  )
  expect_well_formed_pated(pa)
})

# ---------------------------------------------------------------------------
test_that("gee_ works through jointCovariance + pated", {
  long <- sim_repeated(n_subj = 120, n_visits = 3, seed = 5)
  # One row per subject for the auxiliary glm. `z` is a baseline covariate
  # generated independently of arm by sim_repeated(); using it (rather than
  # an aggregate of `y`) keeps the prognostic arm-independent, which is the
  # assumption PATED needs for valid inference.
  aux <- long[!duplicated(long$pid), c("pid", "arm", "z")]

  jc <- jointCovariance(
    gee_(y ~ arm, family = "gaussian", corstr = "exchangeable", data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 2),
    data = list(long, aux)
  )
  expect_well_formed_jc(jc, expected_models = 2)

  pa <- pated(
    gee_(y ~ arm, family = "gaussian", corstr = "exchangeable", data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 2),
    data = list(long, aux)
  )
  expect_well_formed_pated(pa)
})

# ---------------------------------------------------------------------------
test_that("quantile_ works through jointCovariance + pated (bootstrap)", {
  dat <- sim_gaussian(n = 200, seed = 7)
  jc <- jointCovariance(
    quantile_(y ~ arm, probs = c(.25, .5, .75), data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data = list(dat),
    nboot = 30,
    seed  = 11
  )
  expect_s3_class(jc, "jointCovariance")
  # quantile_ contributes 3 parameters (one per probability)
  expect_equal(length(jc$id_map[[1]]), 3L)
  expect_true(all(grepl("^arm_\\d+%$", names(jc$id_map[[1]]))))
  expect_true(all(is.finite(coef(jc))))
  V <- vcov(jc); expect_true(is.matrix(V)); expect_lt(max(abs(V - t(V))), 1e-8)

  pa <- pated(
    quantile_(y ~ arm, probs = c(.25, .5, .75), data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 1),
    data  = list(dat),
    nboot = 30,
    seed  = 11
  )
  expect_well_formed_pated(pa)
})

test_that("quantile_ errors with nboot = 0 (no closed-form score)", {
  dat <- sim_gaussian(n = 100, seed = 8)
  expect_error(
    jointCovariance(
      quantile_(y ~ arm, probs = c(.25, .5, .75), data_index = 1),
      data = list(dat)
    ),
    regexp = "no closed-form score"
  )
})

# ---------------------------------------------------------------------------
test_that("mmrm_ works through jointCovariance + pated", {
  skip_if_not_installed("mmrm")
  long <- sim_repeated(n_subj = 120, n_visits = 3, seed = 6)
  aux  <- long[!duplicated(long$pid), c("pid", "arm", "z")]

  jc <- jointCovariance(
    mmrm_(y ~ arm + us(visit | pid), reml = TRUE, data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 2),
    data = list(long, aux)
  )
  expect_well_formed_jc(jc, expected_models = 2)

  pa <- pated(
    mmrm_(y ~ arm + us(visit | pid), reml = TRUE, data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 2),
    data = list(long, aux)
  )
  expect_well_formed_pated(pa)
})

FisherInformation <- function(score1, score2){

  shared_ids <- intersect(rownames(score1), rownames(score2))
  if(length(shared_ids) == 0){
    ret <- matrix(0, nrow = ncol(score1), ncol = ncol(score2))
  }else{
    ret <- cov(score1[shared_ids, , drop = FALSE], score2[shared_ids, , drop = FALSE]) * 
      (length(shared_ids)-1) / length(shared_ids)
  }

  attr(ret, 'n') <- length(shared_ids)

  invisible(ret)
}

#' Generate bootstrap dataset. 
#' When missing data presents, dataset is split into groups. Patients in 
#' the same group have missing on the same covariates. Bootstrap is carried 
#' out in each of the groups
#' @importFrom dplyr summarise
#' @importFrom dplyr group_by
#' 
#' @param data a data frame to be sampled from.
#'
#' @keywords internal
#' @noRd
sampleWithReplacement <- function(data){
  
  uid <- 
    lapply(
      seq_along(data), 
      function(idx){
        df <- data[[idx]]
        df$index <- idx
        df[, c('pid', 'index'), drop = FALSE]
      })
  uid <- do.call(rbind, uid) %>% 
    group_by(.data$pid) %>% 
    summarise(group = paste0(paste0('data-', sort(unique(.data$index))), collapse = ','))
  
  unique_dataset_group <- sort(unique(uid$group))
  selected_uid <- list()
  for(i in seq_along(unique_dataset_group)){
    group <- unique_dataset_group[i]
    ids <- uid$pid[uid$group == group]
    selected_uid[[group]] <- sample(ids, size = length(ids), replace = TRUE)
  }
  
  bdata <- rep(list(NULL), length(data))
  for(group in names(selected_uid)){
    dat_ids <- as.integer(unlist(strsplit(gsub('data-', '', group), ',')))
    for(dat_id in dat_ids){
      if(!any(duplicated(data[[dat_id]]$pid))){
        rownames(data[[dat_id]]) <- data[[dat_id]]$pid
        tmp_data <- data[[dat_id]][selected_uid[[group]], , drop = FALSE]
        tmp_data$pid <- rownames(tmp_data)
        bdata[[dat_id]] <- rbind(bdata[[dat_id]], tmp_data)
        rm(tmp_data)
      }else{
        tmp <- data[[dat_id]][!duplicated(data[[dat_id]]$pid), , drop = FALSE]
        rownames(tmp) <- tmp$pid
        tmp <- tmp[selected_uid[[group]], 'pid', drop = FALSE]
        for(i in seq_along(selected_uid[[group]])){
          sid <- selected_uid[[group]][i]
          new_sid <- rownames(tmp)[i]
          tmp_data <- data[[dat_id]] %>% dplyr::filter(.data$pid %in% sid)
          tmp_data$pid <- new_sid
          bdata[[dat_id]] <- rbind(bdata[[dat_id]], tmp_data)
          rm(tmp_data, sid, new_sid)
        }
      }
    }
  }
  
  bdata
  
}

#' 
#' @export
jointCovariance.default <- function(..., data, nboot = 0, compute_cov = TRUE, seed = NULL){

  stopifnot(is.list(data))
  if(is.data.frame(data)){
    data <- list(data)
  }
  for(i in 1:length(data)){
    stopifnot(is.data.frame(data[[i]]))
    if(!('pid' %in% names(data[[i]]))){
      stop('data[[', i, ']] does not consist of a column pid for subject IDs. ')
    }
    data[[i]]$pid <- paste0('PID-', data[[i]]$pid)
  }

  specs <- list(...)
  for (i in seq_along(specs)) {
    di <- specs[[i]]$data_index
    if (!is.null(di) && di > length(data)) {
      stop(sprintf(
        "Model %d uses data_index = %d but only %d data frame(s) were supplied.",
        i, di, length(data)), call. = FALSE)
    }
  }
  
  if(nboot > 0){
    fit <- bootstrapJointCovariance(..., data = data, nboot = nboot, compute_cov = compute_cov, seed = seed)
  }else{
    fit <- asymptoticJointCovariance(..., data = data)
  }
  
  fit$call <- match.call()
  class(fit) <- 'jointCovariance'
  invisible(fit)
  
}

# data_index validation, exercised at two layers:
#   1. validate_data_index() called inside each *_() constructor
#   2. a bounds check (data_index <= length(data)) inside jointCovariance.default()

# ---------------------------------------------------------------------------
test_that("validate_data_index() rejects bad inputs", {
  bad_inputs <- list(
    "NA"             = NA_integer_,
    "NA_real"        = NA_real_,
    "NULL"           = NULL,
    "zero"           = 0,
    "negative"       = -1L,
    "non-integer"    = 1.5,
    "string"         = "foo",
    "vector-length2" = c(1L, 2L),
    "logical"        = TRUE
  )
  for (lab in names(bad_inputs)) {
    expect_error(multipleOutcomes:::validate_data_index(bad_inputs[[lab]]),
                 regexp  = "positive integer scalar",
                 label   = lab)
  }
})

test_that("validate_data_index() accepts valid inputs", {
  expect_equal(multipleOutcomes:::validate_data_index(1L), 1L)
  expect_equal(multipleOutcomes:::validate_data_index(1),  1L)
  expect_equal(multipleOutcomes:::validate_data_index(2.0), 2L)
  expect_equal(multipleOutcomes:::validate_data_index(7L), 7L)
})

# ---------------------------------------------------------------------------
test_that("every *_() constructor calls validate_data_index()", {
  # Each constructor should error on data_index = -1 with the validator message.
  expect_error(glm_(y ~ arm, family = "gaussian", data_index = -1),
               "positive integer scalar")
  expect_error(coxph_(Surv(time, event) ~ arm, data_index = 0),
               "positive integer scalar")
  expect_error(logrank_(Surv(time, event) ~ arm, data_index = 1.7),
               "positive integer scalar")
  expect_error(gee_(y ~ arm, family = "gaussian", corstr = "exchangeable",
                    data_index = NA),
               "positive integer scalar")
  expect_error(km_(Surv(time, event) ~ arm, conf_type = "log",
                   data_index = "foo"),
               "positive integer scalar")
  expect_error(quantile_(y ~ arm, probs = c(.25, .5),
                         data_index = c(1L, 2L)),
               "positive integer scalar")
  # mmrm_ goes through additional validation (control checks) before reaching
  # the data_index check, so just verify the error still surfaces. mmrm is in
  # Imports so it's safe to require here.
  expect_error(mmrm_(y ~ arm + us(visit | pid), reml = TRUE, data_index = -3),
               "positive integer scalar")
})

# ---------------------------------------------------------------------------
test_that("jointCovariance() rejects data_index that exceeds length(data)", {
  set.seed(1)
  n <- 50
  d1 <- data.frame(pid = paste0("p", seq_len(n)),
                   arm = rbinom(n, 1, 0.5), y = rnorm(n))

  # One data frame supplied; spec asks for index 2.
  expect_error(
    jointCovariance(
      glm_(y ~ arm, family = "gaussian", data_index = 2L),
      data = list(d1)
    ),
    regexp = "Model 1 uses data_index = 2 but only 1 data frame"
  )

  # Two data frames supplied; second spec asks for index 3.
  d2 <- d1
  expect_error(
    jointCovariance(
      glm_(y ~ arm, family = "gaussian", data_index = 1L),
      glm_(y ~ arm, family = "gaussian", data_index = 3L),
      data = list(d1, d2)
    ),
    regexp = "Model 2 uses data_index = 3 but only 2 data frame"
  )
})

# ---------------------------------------------------------------------------
test_that("multi-data-frame setup passes validation", {
  set.seed(3)
  n <- 60
  # Two data frames keyed on a shared `pid` column, both passing the bounds
  # check and the pid-presence check. data_index = 2 on the auxiliary glm is
  # the case validate_data_index/bounds-check should accept.
  d1 <- data.frame(pid = paste0("p", seq_len(n)),
                   arm = rbinom(n, 1, 0.5), y = rnorm(n))
  d2 <- data.frame(pid = paste0("p", seq_len(n)),
                   arm = rbinom(n, 1, 0.5), z = rnorm(n))
  fit <- jointCovariance(
    glm_(y ~ arm, family = "gaussian", data_index = 1),
    glm_(z ~ arm, family = "gaussian", data_index = 2),
    data = list(d1, d2)
  )
  expect_s3_class(fit, "jointCovariance")
  expect_equal(length(fit$id_map), 2L)
  expect_equal(dim(fit$n_shared_sample_sizes), c(2L, 2L))
  # Both rows of pids match exactly, so the cross-block share count == n.
  expect_equal(fit$n_shared_sample_sizes[1, 2], n)
})

# ---------------------------------------------------------------------------
test_that("default data_index = 1 still works (no explicit arg)", {
  set.seed(2)
  n <- 50
  d <- data.frame(pid = paste0("p", seq_len(n)),
                  arm = rbinom(n, 1, 0.5), y = rnorm(n))

  # No data_index, single data.frame, no list() wrap — the shortest form.
  fit <- jointCovariance(
    glm_(y ~ arm, family = "gaussian"),
    data = d
  )
  expect_s3_class(fit, "jointCovariance")
  expect_equal(length(coef(fit)), 2L)
})

# Shared helpers for the Monte Carlo covariance tests
# (test-mc-covariance.R, test-mc-cross-engine.R). Loaded automatically by
# testthat from any helper-*.R file under tests/testthat/.

# Gate: skip MC tests unless explicitly opted in via env var.
# skip_on_cran() already handles the CRAN case; this lets contributors run
# `devtools::test()` without the slow MC tier by default.
mc_enabled <- function() {
  v <- Sys.getenv("MULTIPLEOUTCOMES_RUN_MC", unset = "")
  isTRUE(v %in% c("1", "true", "TRUE", "yes"))
}

# Run K replicates of a jointCovariance() spec on freshly-simulated data and
# return:
#   beta_hat  : K x p matrix of stacked coefficient vectors
#   vcov_mean : p x p mean of the per-replicate vcov(fit) (theoretical)
#
# `simulate` is a function of one integer arg (the replicate index) returning
# the dataset (either a data.frame or a list, whatever `fit_fn` expects).
# `fit_fn` takes that dataset and returns a jointCovariance object.
# If `model_index` is non-NULL, only that model's coefficients/cov-block are
# kept.
mc_run <- function(K, simulate, fit_fn, model_index = NULL) {
  beta_list <- vector("list", K)
  vcov_list <- vector("list", K)
  for (k in seq_len(K)) {
    dat <- simulate(k)
    fit <- fit_fn(dat)
    if (is.null(model_index)) {
      beta_list[[k]] <- coef(fit)
      vcov_list[[k]] <- vcov(fit)
    } else {
      ids <- fit$id_map[[model_index]]
      beta_list[[k]] <- coef(fit)[ids]
      vcov_list[[k]] <- vcov(fit)[ids, ids, drop = FALSE]
    }
  }
  list(beta_hat  = do.call(rbind, beta_list),
       vcov_mean = Reduce("+", vcov_list) / K)
}

# Assert that an empirical covariance matrix is close to a theoretical one,
# with a relative tolerance on diagonals (variance terms) and an absolute
# tolerance on off-diagonals (covariance terms).
expect_cov_close <- function(emp, thr,
                             rel_diag    = 0.30,
                             abs_offdiag = 0.05,
                             label       = "") {
  d_e <- diag(emp); d_t <- diag(thr)
  rel <- abs(d_e - d_t) / pmax(abs(d_t), 1e-8)
  expect_lt(max(rel),
            rel_diag,
            label = paste("diagonal rel-error", label))

  offdiag_e <- emp; diag(offdiag_e) <- 0
  offdiag_t <- thr; diag(offdiag_t) <- 0
  expect_lt(max(abs(offdiag_e - offdiag_t)),
            abs_offdiag,
            label = paste("off-diagonal abs-error", label))
}

# Monte Carlo validation of the *cross-engine* block of the joint covariance
# matrix. The single-engine and same-engine-pair tests live in
# test-mc-covariance.R; here we explicitly target the diagonal blocks of
# different engines AND the off-diagonal blocks between them.
#
# The kitchen-sink test uses sim_correlated_multi() to draw four endpoints
# (continuous, binary, survival, longitudinal) from a shared latent factor
# vector with exchangeable correlation, so the off-diagonal cross-engine
# blocks of the population covariance are non-zero by construction; if
# jointCovariance() silently set them to zero, the test would catch it via
# both the diagonal/off-diagonal closeness check AND the additional
# non-trivial-correlation sanity check at the end of each test.

# mc_enabled(), mc_run(), expect_cov_close() are in helper-mc.R.

# Sanity check: with rho > 0 in the generator the cross-model correlations
# must be detectable above MC noise.
expect_nontrivial_cross_cor <- function(emp_cov, label = "") {
  cor_mat <- cov2cor(emp_cov)
  offdiag <- cor_mat[upper.tri(cor_mat)]
  expect_gt(max(abs(offdiag)), 0.05,
            label = paste("non-trivial cross-model correlation present", label))
}

# ---------------------------------------------------------------------------
test_that("MC: gee + glm cross-engine covariance is calibrated", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  K <- 200; n <- 250

  res <- mc_run(
    K,
    simulate = function(k) sim_correlated_multi(n_subj = n, seed = 9000L + k),
    fit_fn   = function(dat) jointCovariance(
      gee_(y_long ~ arm, family = "gaussian",
           corstr = "exchangeable", data_index = 1),
      glm_(y_cont ~ arm, family = "gaussian", data_index = 2),
      data = list(dat$long, dat$wide)
    )
  )
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.30, abs_offdiag = 0.04,
                   label = "gee x glm")
  expect_nontrivial_cross_cor(cov(res$beta_hat), label = "(gee x glm)")
})

# ---------------------------------------------------------------------------
test_that("MC: mmrm + glm cross-engine covariance is calibrated", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  skip_if_not_installed("mmrm")
  K <- 200; n <- 250

  res <- mc_run(
    K,
    simulate = function(k) sim_correlated_multi(n_subj = n, seed = 10000L + k),
    fit_fn   = function(dat) jointCovariance(
      mmrm_(y_long ~ arm + us(visit | pid), reml = TRUE, data_index = 1),
      glm_(y_cont ~ arm, family = "gaussian", data_index = 2),
      data = list(dat$long, dat$wide)
    )
  )
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.30, abs_offdiag = 0.04,
                   label = "mmrm x glm")
  expect_nontrivial_cross_cor(cov(res$beta_hat), label = "(mmrm x glm)")
})

# ---------------------------------------------------------------------------
test_that("MC: gee + glm cross-engine covariance with partial overlap (70%)", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  K <- 200; n <- 300

  res <- mc_run(
    K,
    simulate = function(k) sim_correlated_multi(n_subj = n,
                                                long_overlap_frac = 0.7,
                                                seed = 9500L + k),
    fit_fn   = function(dat) jointCovariance(
      gee_(y_long ~ arm, family = "gaussian",
           corstr = "exchangeable", data_index = 1),
      glm_(y_cont ~ arm, family = "gaussian", data_index = 2),
      data = list(dat$long, dat$wide)
    )
  )
  # Tolerance is slightly looser here than the full-overlap case because the
  # cross-block scales by ~0.7 and MC noise scales by 1/sqrt(K_effective).
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.30, abs_offdiag = 0.05,
                   label = "gee x glm (70% overlap)")
  expect_nontrivial_cross_cor(cov(res$beta_hat),
                              label = "(gee x glm partial overlap)")
})

# ---------------------------------------------------------------------------
# Kitchen-sink: 4 engines, 4 endpoint types, correlated outcomes.
test_that("MC: kitchen-sink (glm-gaussian, glm-binomial, coxph, gee) is calibrated", {
  skip_on_cran()
  skip_if_not(mc_enabled(),
              "set MULTIPLEOUTCOMES_RUN_MC=1 to run Monte Carlo tests")
  skip_if_not_installed("survival")
  library(survival)
  K <- 200; n <- 250

  res <- mc_run(
    K,
    simulate = function(k) sim_correlated_multi(n_subj = n, seed = 11000L + k),
    fit_fn   = function(dat) jointCovariance(
      glm_(y_cont ~ arm,                family = "gaussian", data_index = 1),
      glm_(y_bin  ~ arm,                family = "binomial", data_index = 1),
      coxph_(Surv(time, event) ~ arm,                         data_index = 1),
      gee_(y_long ~ arm, family = "gaussian",
           corstr = "exchangeable",                            data_index = 2),
      data = list(dat$wide, dat$long)
    )
  )
  expect_cov_close(cov(res$beta_hat), res$vcov_mean,
                   rel_diag = 0.30, abs_offdiag = 0.05,
                   label = "kitchen-sink")
  expect_nontrivial_cross_cor(cov(res$beta_hat), label = "(kitchen-sink)")
})

#' Creating Objects of Group Quantile Differences
#'
#' @description
#' `quantile_` is a wrapper function creating an object that, for each
#' requested probability, computes the difference between the two arms'
#' sample quantiles of the outcome. The object is passed to `jointCovariance`
#' or `pated` through `...`. Because the empirical-quantile estimator has no
#' tractable closed-form score, this engine is available only through the
#' bootstrap path (`nboot > 0`).
#'
#' @param formula a two-sided formula `y ~ arm` where the right-hand side is a
#'   binary grouping variable.
#' @param probs numeric vector of probabilities in `(0, 1)`. By default the
#'   first quartile, median, and third quartile.
#' @param data_index integer. Index of the data frame in the `data` argument
#'   of `jointCovariance` to be used.
#'
#' @export
quantile_ <- function(formula, probs = c(0.25, 0.5, 0.75), data_index = 1) {

  if (!inherits(formula, "formula")) {
    stop('"formula" in quantile_ must be a R formula.')
  }
  if (!is.numeric(probs) || any(probs <= 0 | probs >= 1)) {
    stop('"probs" must be numeric values in (0, 1).')
  }
  data_index <- validate_data_index(data_index)

  structure(
    list(
      engine    = "quantile",
      formula   = formula,
      probs     = probs,
      data_index = data_index,
      id_col    = "pid"
    ),
    class = c("jc_spec_quantile", "jc_spec")
  )
}

# Generate two curves of survival probability with pointwise 95% confidence 
# interval: 
# 1. PATED adjusted KM curve
# 2. Conventional KM curve
comparePointwiseConfidenceIntervalWidth <- function(pated_res, km_res, transform = 'identity'){

  pated_estimate <- pated_res$estimate
  names(pated_estimate) <- rownames(pated_res)
  pated_coordinate <- 
    data.frame(
      Time = extractKaplanMeierTimes(pated_estimate, sort = FALSE),
      Strata = pated_estimate %>% 
        names() %>%
        sapply(function(x) {sub('.*time_\\((.*)\\)_\\(.*\\)', '\\1', x)}),
      pated_estimate = pated_estimate,
      pated_ci_width = pated_res$UCI - pated_res$LCI
    )
  rownames(pated_coordinate) <- NULL
  
  km_estimate <- km_res$estimate
  names(km_estimate) <- rownames(km_res)
  km_coordinate <- 
    data.frame(
      Time = extractKaplanMeierTimes(km_estimate, sort = FALSE),
      Strata = km_estimate %>% 
        names() %>%
        sapply(function(x) {sub('.*time_\\((.*)\\)_\\(.*\\)', '\\1', x)}),
      km_estimate = km_estimate,
      km_ci_width = km_res$UCI - km_res$LCI
    )
  rownames(km_coordinate) <- NULL
  
  m <- merge(pated_coordinate, km_coordinate, by = c('Time', 'Strata'), all = TRUE)
  
  p1 <- 
    ggplot(m, aes(x = .data$Time, y = .data$km_ci_width / .data$pated_ci_width, group = .data$Strata)) + 
      geom_point(aes(color = .data$Strata), size = .5) + 
      scale_x_continuous(transform = transform) + 
      theme_minimal() +
      labs(title = "Widths of Pointwise 95% CI",
           x = str_glue('Time{ifelse(transform == "identity", "", paste(" (", transform, ")"))}'),
           y = 'KM / PATED')
  
  p2 <- 
    ggplot(m, aes(x = .data$km_estimate, y = .data$pated_estimate, group = .data$Strata)) + 
    geom_point(aes(color = .data$Strata), size = .5) + 
    geom_abline(intercept = 0, slope = 1, linetype = 'dashed') + 
    theme_minimal() + 
    labs(title = 'Estimates of S(t)', 
         x = 'KM', 
         y = 'PATED')
  
  list(p1, p2)
}

# Inversed function of g(S(t)).
# See https://documentation.sas.com/doc/en/pgmsascdc/9.4_3.4/statug/statug_lifetest_details08.htm
gInverseFunction <- function(conf.type = c('log', 'log-log', 'plain', 'logit', 'arcsin')){

  conf.type <- match.arg(conf.type)

  switch(conf.type,
    'log'     = function(x) exp(x),
    'log-log' = function(x) exp(-exp(x)),
    'plain'   = function(x) x,
    'logit'   = function(x) exp(x) / (1 + exp(x)),
    'arcsin'  = function(x) sin(pmin(pmax(x, 0), pi / 2))^2
  )
}

# gFunction / gInverseFunction must be exact mutual inverses on the unit
# interval for every documented conf_type, including arcsin.

test_that("g and g-inverse round-trip for every supported conf_type", {
  conf_types <- c("log", "log-log", "plain", "logit", "arcsin")
  x <- seq(0.05, 0.95, by = 0.05)
  for (ct in conf_types) {
    g    <- multipleOutcomes:::gFunction(ct)
    ginv <- multipleOutcomes:::gInverseFunction(ct)
    expect_equal(ginv(g(x)), x, tolerance = 1e-10,
                 label = paste0("ginv(g(x)) for conf_type=", ct))
  }
})

test_that("invalid conf_type is rejected by g/g-inverse", {
  expect_error(multipleOutcomes:::gFunction("oops"))
  expect_error(multipleOutcomes:::gInverseFunction("oops"))
})

MMRMAdapter <- R6::R6Class(
  "MMRMAdapter",
  inherit = BaseAdapter,
  public = list(
    fit_model = function(){
      
      self$type <- 'mmrm'
      
      self$data[[self$spec$id_col]] <- as.factor(self$data[[self$spec$id_col]])
      
      self$fit <- mmrm::mmrm(formula = self$spec$formula, 
                             data = self$data, 
                             covariance = self$spec$covariance,
                             reml = self$spec$reml, 
                             control = self$spec$control)
      
      ## estimate
      self$estimate <- stats::coef(self$fit)
      
      ## score
      tr <- try(self$score <- mmrm::component(self$fit, 'score_per_subject'), silent = TRUE)
      if('try-error' %in% class(tr)){
        stop('score_per_subject is not available. Check mmrm version installed in your session. Contact author for assistance. ')
      }
      
      ## It is very tricky to align rows in score_per_subject and sample ID.
      ## I don't know details of how mmrm is implemented. 
      ## For example, the input data frame does not 
      ## assume samples from a subject are physically contiguous records. 
      ## Also, records with NA may be discarded. As a result, it is unknown 
      ## how to map rows in score with subjects.
      ## I use the following approach as a guardrail: 
      ## 1. extract model matrix from fitted object where only samples being 
      ##    used in fitted the model are retained. This is a record-wise 
      ##    data frame where a subject may have data in more than one row. 
      ##    See ?stats::model.matrix. This matrix does not have the pid column. 
      ## 2. extract full_frame from fitted object. This data frame consists 
      ##    of all columns, including pid. See ?mmrm::component
      ## 3. For any variable that is in both data frames extracted in step 1 and 
      ##    step 2, check whether they are identical. With moderate or large 
      ##    sample size, it is unlikely that all variables are identical in 
      ##    two data frames if they are not well-aligned. 
      ## 4. If the test in step 3 passes, I assume that the orders of pid in 
      ##    score and full_frame are the same. I need to request for 
      ##    confirmation from mmrm authors. 
      
      m1 <- as.data.frame(stats::model.matrix(self$fit))
      m2 <- mmrm::component(self$fit, 'full_frame')
      
      shared_cols <- intersect(names(m1), names(m2))
      
      for(col in shared_cols){
        if(max(abs(m1[, col] - m2[, col])) > 1e-4){
          stop(col, ' does not match. Unable to map score to subject ID. ')
        }
      }
      
      ids <- m2[[self$spec$id_col]]
      rownames(self$score) <- ids[which(!duplicated(ids))]
      self$sample_id <- rownames(self$score)
      
      stopifnot(nrow(self$score) == mmrm::component(self$fit, 'n_subjects'))
      self$n <- mmrm::component(self$fit, 'n_subjects')
      
      ## inv_hess
      self$inv_hess <- -self$fit$beta_vcov * self$n
      
      invisible(self)
    },
    
    refit = function(bdata){
      bdata[[self$spec$id_col]] <- as.factor(bdata[[self$spec$id_col]])
      bfit <- mmrm::mmrm(formula = self$spec$formula, 
                         data = bdata, 
                         covariance = self$spec$covariance,
                         reml = self$spec$reml, 
                         control = self$spec$control)
      stats::coef(bfit)
    }
    
  )
)

CoxphAdapter <- R6::R6Class(
  "CoxphAdapter",
  inherit = BaseAdapter,
  public = list(
    fit_model = function(){
      
      self$type <- 'coxph'
      
      self$fit <- survival::coxph(formula = self$spec$formula, 
                                  data = self$data, model = TRUE)
      
      ## estimate
      self$estimate <- stats::coef(self$fit)
      
      ## score
      self$score <- -as.matrix(resid(self$fit, type = 'score'))
      ## unlike in glm, stats::nobs(self$fit) would return the number of 
      ## samples being used in model fitting. For coxph, stats::nobs(self$fit) 
      ## returns number of events, which is not what we need here
      self$n <- nrow(self$score)
      
      kept_ids <- setdiff(seq_len(nrow(self$data)), self$fit$na.action)
      rownames(self$score) <- self$data[[self$spec$id_col]][kept_ids]
      self$sample_id <- rownames(self$score)
      
      ## inv_hess
      self$inv_hess <- solve(imatCoxph(self$fit))
      
      invisible(self)
    },
    
    refit = function(bdata) {
      bfit <- survival::coxph(formula = self$spec$formula, 
                              data = bdata)
      stats::coef(bfit)
    }
  )
)

# Figure out the time points with at least one event. 
# This is used to generate KM-like curves.
extractKaplanMeierTimes <- function(coef, sort = TRUE){
  
  labels <- names(coef)
  times <- labels %>% 
    sapply(function(x) {sub('.*time_\\(.*\\)_\\((.*)\\)', '\\1', x)}) %>% 
    as.numeric() %>% 
    unique()
  
  if(sort){
    times <- sort(times)
  }
  
  times
  
}

# Validate that data_index is a single positive integer (or integer-valued
# double). Returns the value coerced to integer. Called from every *_()
# constructor; the bounds check (data_index <= length(data)) lives later in
# jointCovariance.default() because that's the first place we know how many
# data frames were supplied.
validate_data_index <- function(data_index) {
  if (!is.numeric(data_index) ||
      length(data_index) != 1L ||
      is.na(data_index) ||
      data_index < 1 ||
      data_index != as.integer(data_index)) {
    stop("`data_index` must be a positive integer scalar; got ",
         deparse(data_index), call. = FALSE)
  }
  as.integer(data_index)
}

#' Prognostic Variables Assisted Treatment Effect Detection
#' @description
#' `pated` is a wrapper function of `jointCovariance` for testing treatment effect
#' in randomized clinical trials. It assumes that prognostic variables are fully
#' randomized. This assumption can help enhancing statistical power of conventional
#' approaches in detecting the treatment effect. Specifically, the sensitivity
#' of the conventional models specified in `...` are improved by `pated`.
#' 
#' @importFrom stringr str_extract
#' @importFrom ggpubr ggarrange
#' @importFrom stats coef cov cov2cor pchisq printCoefmat rbinom rnorm runif sd time var vcov
#' @importFrom dplyr row_number first case_when bind_rows filter arrange desc 
#' @importFrom ggplot2 geom_hline geom_vline scale_color_manual theme element_blank
#' @importFrom rlang .data
#' 
#' @param ... model specifications built by `glm_()`, `coxph_()`, `logrank_()`,
#' `gee_()`, `mmrm_()`, `km_()`, or `quantile_()`. The first specification is
#' the primary outcome whose treatment effect is being tested; the rest are
#' prognostic covariates used to tighten the SE.
#' @param data either a single data frame (when all models are fitted on
#' the same dataset) or a list of data frames (one entry per `data_index`).
#' Each data frame must have a `pid` column carrying subject identifiers;
#' records with the same `pid` across different data frames refer to the
#' same subject.
#' @param nboot non-zero integer if bootstrap is adopted. By default 0.
#' @param compute_cov logic. If \code{TRUE}, empirical covariance matrix is computed 
#' using bootstrap estimate and returned. Bootstrap estimate will be abandoned. If 
#' \code{FALSE}, bootstrap estimate will be returned and no empirical covariance 
#' matrix is computed. 
#' @param seed random seed when generate bootstrap data. 
#' @param transform character. Now only supports \code{"identity"}. 
#'
#' @return a data frame of testing results.
#' @export
#'
#' @examples
#' ## More examples can be found in the vignettes.
#' library(survival)
#' library(mvtnorm)
#'
#' genData <- function(seed = NULL){
#'   set.seed(seed)
#'   n <- 300
#'   sigma <- matrix(c(1, .6, .6, 1), 2)
#'   x <- rmvnorm(n, sigma = sigma)
#'   z1 <- rbinom(n, 1, .6)
#'   z2 <- rnorm(n)
#'   trt <- rbinom(n, 1, .5)
#'
#'   bet <- c(-.2, .2)
#'   y <- -.5 + x %*% bet + z1 * .3 - z2 * .1 + .1 * trt - .1 * rnorm(n)
#'   death <- rbinom(n, 1, .8)
#'   data.frame(
#'     y = as.numeric(y), trt = trt,
#'     z1 = z1, z2 = z2,
#'     x1 = x[, 1], x2 = x[, 2],
#'     death, pid = paste0('s-', seq_len(n))
#'   )
#' }
#'
#' dat <- genData(seed = 31415926)
#'
#' ## `data_index` defaults to 1 in every spec constructor and a single
#' ## data.frame is auto-wrapped into a list, so neither needs spelling out
#' ## when all models are fitted on the same dataset.
#' fit <-
#'   pated(
#'     coxph_(Surv(time = y, event = death) ~ trt),
#'     glm_(z1 ~ trt, family = 'binomial'),
#'     glm_(z2 ~ trt, family = 'gaussian'),
#'     glm_(x1 ~ trt, family = 'gaussian'),
#'     glm_(x2 ~ trt, family = 'gaussian'),
#'     data = dat
#'   )
#'
#' fit
#'

pated <- 
  function(
    ..., 
    data, 
    nboot = 0, 
    compute_cov = TRUE, 
    seed = NULL, 
    transform = 'identity'){
    
  fit <- 
    jointCovariance(
      ..., 
      data = data, 
      nboot = nboot, 
      compute_cov = compute_cov, 
      seed = seed)

  fml <- parseTreatmentVariableFromCall(...)

  id <- 
    lapply(
      seq_along(fit$id_map), 
      function(i){
        imap <- fit$id_map[[i]]
        arm <- fml$arm[i]
        if(arm %in% names(imap)){
          imap[arm]
        }else{
          if(any(grepl(str_glue('^{arm}_\\d+%$'), names(imap)))){
            imap[grepl(str_glue('^{arm}_\\d+%$'), names(imap))]
          }else{
            imap[startsWith(str_extract(names(imap), "(?<=time_\\().*?(?=\\)_\\()"), arm)]
          }
        }
      }
    )
  id1 <- id[[1]]
  id2 <- unlist(id[-1])
  Delta <- coef(fit)[id1]
  delta <- coef(fit)[id2]
  mcov <- vcov(fit)
  if(!is.null(mcov)){
    mcov22 <- mcov[id2, id2, drop = FALSE]
    mcov21 <- mcov[id2, id1, drop = FALSE]
    var11 <- diag(mcov[id1, id1, drop = FALSE])
  }else{
    mcov22 <- cov(fit$bootstrap_estimate[, id2, drop = FALSE])
    mcov21 <- cov(fit$bootstrap_estimate[, id2, drop = FALSE], fit$bootstrap_estimate[, id1, drop = FALSE])
    var11 <- apply(fit$bootstrap_estimate[, id1, drop = FALSE], 2, var)
    if(!('km_' %in% fml$func)){
      mcov <- cov(fit$bootstrap_estimate)
    }
  }
  opt_c <- - solve(mcov22) %*% mcov21

  estimate <- Delta + as.vector(t(opt_c) %*% delta)
  variance <- var11 +
              diag(t(opt_c) %*% mcov21) +
              diag(t(mcov21) %*% opt_c) +
              diag(t(opt_c) %*% mcov22 %*% opt_c)
  if (any(variance < 0, na.rm = TRUE)) {
    warning(
      "PATED residual variance is negative for ",
      sum(variance < 0, na.rm = TRUE),
      " of ", length(variance), " term(s); the resulting stderr / p-value ",
      "will be NaN. This typically means one or more prognostic variables ",
      "are nearly collinear with the primary outcome (or with each other), ",
      "so the adjustment leaves no residual variance to estimate. Check ",
      "that the prognostic variables are independent of arm assignment ",
      "and not redundant with the primary outcome.",
      call. = FALSE
    )
  }
  stderr <- sqrt(variance)

  pvalue <- pchisq((estimate / stderr)^2, df = 1, lower.tail = FALSE)
  
  treatment <-
    data.frame(
      term = fml$outcome[1],
      family = 'PATED',
      estimate = estimate,
      stderr = stderr,
      pvalue = pvalue,
      method = 'PATED',
      corr = NA
    )

  family <- sub('_$', '', fml$func)
  family <- c(rep(family[1], length(id1)), rep(family[-1], times = fml$n_terms[-1]))
  
  nonconfounder <-
    data.frame(
      term = c(rep(fml$outcome[1], length(id1)), rep(fml$outcome[-1], times = fml$n_terms[-1])),
      family = family,
      estimate = coef(fit)[c(id1, id2)],
      stderr = sqrt(c(var11, diag(mcov22))),
      pvalue = pchisq((coef(fit)[c(id1, id2)] / sqrt(c(var11, diag(mcov22))))^2, df = 1, lower.tail = FALSE),
      method = rep(c('Standard', 'Prognostic'), times = c(length(id1), length(id2))),
      corr = NA
    )
  if(!is.null(mcov) && length(id1) == 1){
    nonconfounder$corr <- cov2cor(mcov)[c(id1, id2), id1]
  }

  ret <- rbind(treatment, nonconfounder %>% arrange(desc(abs(.data$corr))))
  ret$method <- paste0(ret$method, ifelse(ret$method == 'Prognostic' & ret$pvalue < .05, '*', ''))
  if(!is.null(mcov) && length(id1) == 1){
    attr(ret, 'Rel. Eff.') <- mcov[id1, id1] / stderr^2
  }
  
  if(fml$func[1] %in% 'km_'){

    conf.type <- fml$arg[fml$func %in% 'km_'][1]
    pated_res <- 
      treatment %>% 
      dplyr::select(estimate, stderr) %>% 
      mutate(LCI = gInverseFunction(conf.type)(estimate - 1.96 * stderr)) %>%
      mutate(UCI = gInverseFunction(conf.type)(estimate + 1.96 * stderr)) %>% 
      mutate(estimate = gInverseFunction(conf.type)(estimate)) %>% 
      dplyr::select(dplyr::all_of(c("estimate", "LCI", "UCI")))
    
    km_res <- 
      nonconfounder %>% 
      dplyr::filter(.data$method %in% 'Standard') %>% 
      dplyr::select(estimate, stderr) %>% 
      mutate(LCI = gInverseFunction(conf.type)(estimate - 1.96 * stderr)) %>%
      mutate(UCI = gInverseFunction(conf.type)(estimate + 1.96 * stderr)) %>% 
      mutate(estimate = gInverseFunction(conf.type)(estimate)) %>% 
      dplyr::select(dplyr::all_of(c("estimate", "LCI", "UCI")))
    
    pated_curve <- createKaplanMeierCurve(pated_res, str_glue('PATED ({conf.type})'), transform)
    km_curve <- createKaplanMeierCurve(km_res, str_glue('KM ({conf.type})'), transform)
    comp_ci <- comparePointwiseConfidenceIntervalWidth(pated_res, km_res, transform)
    
    #plot(ggarrange(ggarrange(km_curve, pated_curve, ncol = 2), ggarrange(comp_ci[[1]], comp_ci[[2]], ncol = 2), nrow = 2))
    plot(ggarrange(km_curve, pated_curve, ncol = 2))
    
    attr(ret, 'pated curve') <- pated_curve
    attr(ret, 'km curve') <- km_curve
    attr(ret, 'comp ci') <- comp_ci
    attr(ret, 'conf.type') <- conf.type
  }
  class(ret) <- c('pated', class(ret))
  ret

}

#' Plot PATED Analysis Results
#'
#' @param x an object returned from \code{pated()}. 
#' @param ... currently not supported.
#' 
#' @returns \code{NULL}
#' @export
#'
plot.pated <- function(x, ...){
  
  x_plot <- x %>%
    mutate(term = ifelse(row_number() == 2 & .data$term == first(.data$term),
                         paste0(" ", .data$term),
                         .data$term)) %>%
    mutate(
      rowid = row_number(),
      z = .data$estimate / .data$stderr,
      group = case_when(
        rowid == 1 ~ "PATED",
        rowid == 2 ~ "Unadjusted",
        TRUE ~ "Prognostic"
      )
    )
  
  x_plot <- bind_rows(
    x_plot %>% filter(.data$rowid <= 2),
    x_plot %>% filter(.data$rowid > 2) %>% arrange(desc(abs(.data$corr)))
  )
  
  x_plot$variable <- factor(x_plot$term, levels = rev(x_plot$term))
  
  p <- 
    ggplot(x_plot, aes(x = .data$z, y = .data$variable, color = .data$group)) +
    geom_point(size = 3) +
    geom_hline(yintercept = length(x_plot$variable) - 2 + 0.5,
               linetype = "dashed", color = "gray80") +
    geom_vline(xintercept = 0, linetype = "dashed") +
    geom_vline(xintercept = c(-1.96, 1.96), linetype = "dashed", color = "gray60") +
    scale_color_manual(values = c("PATED" = "red",
                                  "Unadjusted" = "blue",
                                  "Prognostic" = "gray40"),
                       breaks = c("PATED", "Unadjusted", "Prognostic")) +
    labs(x = "Z value (variable ~ arm)", y = "Variable") +
    theme_minimal() +
    theme(legend.title = element_blank())
  
  plot(p)
  
  invisible(NULL)
  
}

# Regression fixture for the README example. Locks in the deterministic
# (asymptotic) pated() output on the bundled `indo` dataset. If a future
# change to GLMAdapter, FisherInformation, or pated's adjustment formula
# shifts any number by more than 1e-10, this test fails immediately and
# the README/help-page numbers (which are generated from the same call)
# stop being trustworthy.

test_that("README pated() example reproduces", {
  data(indo, package = "multipleOutcomes")
  indo <- indo %>% dplyr::mutate(pid = paste0("s-", id))

  fit <- pated(
    glm_(outcome     ~ rx, family = "binomial"),
    glm_(risk        ~ rx, family = "gaussian"),
    glm_(gender      ~ rx, family = "binomial"),
    glm_(sod         ~ rx, family = "binomial"),
    glm_(pep         ~ rx, family = "binomial"),
    glm_(recpanc     ~ rx, family = "binomial"),
    glm_(psphinc     ~ rx, family = "binomial"),
    glm_(precut      ~ rx, family = "binomial"),
    glm_(difcan      ~ rx, family = "binomial"),
    glm_(amp         ~ rx, family = "binomial"),
    glm_(paninj      ~ rx, family = "binomial"),
    glm_(acinar      ~ rx, family = "binomial"),
    glm_(asa81       ~ rx, family = "binomial"),
    glm_(asa         ~ rx, family = "binomial"),
    glm_(prophystent ~ rx, family = "binomial"),
    glm_(therastent  ~ rx, family = "binomial"),
    glm_(pdstent     ~ rx, family = "binomial"),
    data = indo
  )

  ref <- readRDS(
    system.file("testdata", "readme_pated_reference.rds",
                package = "multipleOutcomes")
  )

  # Row order: PATED first, Standard second, then prognostic rows. The fixture
  # captured the same ordering.
  expect_equal(as.character(fit$term),    ref$term)
  expect_equal(as.character(fit$family),  ref$family)
  expect_equal(as.character(fit$method),  ref$method)

  expect_equal(unname(as.numeric(fit$estimate)), ref$estimate, tolerance = 1e-10)
  expect_equal(unname(as.numeric(fit$stderr)),   ref$stderr,   tolerance = 1e-10)
  expect_equal(unname(as.numeric(fit$pvalue)),   ref$pvalue,   tolerance = 1e-10)
  expect_equal(unname(as.numeric(fit$corr)),     ref$corr,     tolerance = 1e-10)
  expect_equal(unname(as.numeric(attr(fit, "Rel. Eff."))),
               ref$rel_eff, tolerance = 1e-10)
})

# Regression tests for the km_ + pated() pipeline.
#
# The reference fixture inst/testdata/legacy_km_reference.rds was produced from
# the legacy `kmMO + pated(family=..., data_index=...)` API on the main HEAD of
# this repository, run with the same data generator and the same seed.
#
# Quantities split into two tiers:
#   1. Deterministic (data-only): unadjusted KM transformed-S(t), and GLM
#      coefficients of the prognostic covariates. These do not depend on the
#      bootstrap RNG state. We test them at high precision.
#   2. Bootstrap-driven: PATED estimates and SEs, KM SEs. The legacy and new
#      implementations sample different bootstrap sequences (legacy keyed on
#      integer `id`; new keys on string `pid`), so we only sanity-check
#      orderings and inferential conclusions, not pointwise equality.

ref <- readRDS(
  system.file("testdata", "legacy_km_reference.rds", package = "multipleOutcomes")
)

eval_times <- ref$data_signature$eval_times

dat <- gen_delayed_effect_data(seed = ref$data_signature$seed_data)

test_that("data signature matches the legacy reference", {
  expect_equal(nrow(dat),       ref$data_signature$n)
  expect_equal(sum(dat$event),  ref$data_signature$n_event)
  expect_equal(as.list(table(dat$arm)), ref$data_signature$arm_table)
})

# Single pated() fit reused by all downstream assertions.
fit <- pated(
  km_(Surv(time, event) ~ arm, conf_type = "log",
      times = eval_times, data_index = 1),
  glm_(x1 ~ arm, family = "gaussian", data_index = 1),
  glm_(x2 ~ arm, family = "binomial", data_index = 1),
  glm_(z1 ~ arm, family = "gaussian", data_index = 1),
  data        = list(dat),
  nboot       = ref$data_signature$nboot,
  compute_cov = FALSE,
  seed        = ref$data_signature$seed_boot
)

test_that("pated() returns a well-formed result", {
  expect_s3_class(fit, "pated")
  expect_true(is.data.frame(fit))
  expect_true(all(c("term","family","estimate","stderr","pvalue","method","corr")
                  %in% names(fit)))
  # legacy had 47 rows: 22 PATED rows + 22 unadjusted KM rows + 3 prognostic glm rows
  expect_equal(nrow(fit), nrow(ref$pated_table))
})

test_that("unadjusted KM transformed-S(t) matches legacy exactly", {
  new_km <- fit$estimate[fit$family == "km"]
  expect_equal(length(new_km), length(ref$unadjusted_km))
  # log-transformed S(t) is computed deterministically on the original data,
  # so legacy vs new must agree to numeric precision.
  expect_equal(unname(new_km), unname(ref$unadjusted_km), tolerance = 1e-10)
})

test_that("GLM prognostic coefficients match legacy exactly", {
  prog <- fit$method %in% c("Prognostic", "Prognostic*")
  new_glm <- setNames(fit$estimate[prog], fit$term[prog])
  expect_equal(sort(names(new_glm)), sort(names(ref$glm_estimates)))
  expect_equal(new_glm[names(ref$glm_estimates)],
               ref$glm_estimates, tolerance = 1e-10)
})

test_that("KM curve coordinates agree with legacy on S(t) point estimates", {
  new_curve <- attr(attr(fit, "km curve"), "coordinate")
  m <- merge(new_curve, ref$km_curve_data,
             by = c("Time", "Strata"), suffixes = c(".new", ".old"))
  expect_equal(nrow(m), nrow(ref$km_curve_data))
  # S(t) is gInverse(log-S(t)); both should match to numeric precision.
  expect_equal(m$St.new, m$St.old, tolerance = 1e-10)
  # LCI/UCI use bootstrap SEs and differ across runs, but stay within MC noise.
  expect_lt(max(abs(m$LCI.new - m$LCI.old)), 0.05)
  expect_lt(max(abs(m$UCI.new - m$UCI.old)), 0.05)
})

test_that("PATED treatment-effect estimates are close to legacy", {
  new_pated <- fit$estimate[fit$method == "PATED"]
  old_pated <- ref$pated_table$estimate[ref$pated_table$method == "PATED"]
  expect_equal(length(new_pated), length(old_pated))
  expect_lt(max(abs(new_pated - old_pated)), 0.05)
})

test_that("KMAdapter preserves names on its estimate (regression for (a))", {
  jc <- jointCovariance(
    km_(Surv(time, event) ~ arm, conf_type = "log",
        times = eval_times, data_index = 1),
    glm_(x1 ~ arm, family = "gaussian", data_index = 1),
    data  = list(dat),
    nboot = 5,
    seed  = 1
  )
  expect_false(is.null(names(jc$id_map[[1]])))
  expect_true(all(grepl("^time_\\(arm=[01]\\)_\\(", names(jc$id_map[[1]]))))
})

test_that("inferential conclusions on prognostic covariates are stable", {
  pvals <- setNames(fit$pvalue[fit$method %in% c("Prognostic", "Prognostic*")],
                    fit$term[fit$method %in% c("Prognostic", "Prognostic*")])
  # x2 is the prognostic covariate with the strongest signal in this dataset.
  expect_lt(pvals[["x2"]], 0.10)
  # x1 and z1 are not informative for the survival endpoint here.
  expect_gt(pvals[["x1"]], 0.30)
  expect_gt(pvals[["z1"]], 0.05)
})

# Regression test: fitKMCurve() must not return NA estimates when the
# requested time grid extends past the last observed event in a stratum.
# Previously the function called summary(..., extend = FALSE) and merged with
# NAs, which then propagated through cov() in the bootstrap and broke pated().

test_that("fitKMCurve() carries S(t) forward past last event (no NAs)", {
  skip_if_not_installed("survival")
  library(survival)

  set.seed(99)
  n <- 60
  dat <- data.frame(
    pid  = paste0("p", seq_len(n)),
    arm  = c(rep(0, n/2), rep(1, n/2))
  )
  # arm=0 has all events by t=3; arm=1 has events spread out to t=10. Asking for
  # times past 3 forces the "extend" branch in arm=0.
  dat$time  <- ifelse(dat$arm == 0,
                      runif(n, 0.5, 3),
                      runif(n, 0.5, 10))
  dat$event <- ifelse(dat$arm == 0, 1, rbinom(n, 1, 0.6))

  trans_St <- multipleOutcomes:::fitKMCurve(
    Surv(time, event) ~ arm,
    data      = dat,
    conf_type = "log",
    times     = c(1, 2, 3, 4, 6, 9)
  )

  expect_false(any(is.na(trans_St)))
  # Both strata must have exactly the same set of evaluation times.
  expect_equal(length(trans_St), 2 * 6)
})

# Targeted tests to lift coverage of accessor / printing / utility code paths
# that the engine-specific tests don't naturally exercise.

helper_simple_fit <- function(seed = 13) {
  set.seed(seed)
  n <- 200
  dat <- data.frame(
    pid = paste0("p", seq_len(n)),
    arm = rbinom(n, 1, 0.5),
    z   = rnorm(n)
  )
  dat$y <- 0.3 * dat$arm + 0.5 * dat$z + rnorm(n)
  jointCovariance(
    glm_(y ~ arm, family = "gaussian"),
    glm_(z ~ arm, family = "gaussian"),
    data = dat
  )
}

# ---------------------------------------------------------------------------
test_that("summary(jc) and print(summary(jc)) work", {
  fit <- helper_simple_fit()
  s   <- summary(fit)
  expect_s3_class(s, "summary.jointCovariance")
  expect_true(is.matrix(s$coefficients))
  expect_true(all(c("Estimate", "Std. Error", "z value", "Pr(>|z|)")
                  %in% colnames(s$coefficients)))
  expect_output(print(s), "Estimate")
})

test_that("summary(jc, model_index = N) restricts to that model", {
  fit <- helper_simple_fit()
  s_all <- summary(fit)
  s_one <- summary(fit, model_index = 1)
  expect_lt(nrow(s_one$coefficients), nrow(s_all$coefficients))
  expect_equal(nrow(s_one$coefficients), length(fit$id_map[[1]]))
  expect_error(summary(fit, model_index = 99), "Invalid model index")
})

# ---------------------------------------------------------------------------
test_that("coef(jc, model_index) and vcov(jc, model_index) restrict properly", {
  fit <- helper_simple_fit()
  c1 <- coef(fit, 1)
  v1 <- vcov(fit, 1)
  expect_true(!is.null(names(c1)))
  expect_equal(dim(v1), c(length(c1), length(c1)))
  expect_equal(rownames(v1), names(c1))
  expect_error(coef(fit, 99), "Invalid model index")
  expect_error(vcov(fit, 99), "Invalid model index")
})

test_that("coef(jc) and vcov(jc) (no model_index) return the full vector / matrix", {
  fit <- helper_simple_fit()
  expect_equal(length(coef(fit)), 4L)        # 2 models x (Intercept, arm)
  expect_equal(dim(vcov(fit)),   c(4L, 4L))
})

# ---------------------------------------------------------------------------
test_that("simulateMoData() produces a list of two data frames", {
  d <- simulateMoData(n = 80, hr = 0.7, seed = 1)
  expect_type(d, "list")
  expect_true(all(c("dat1", "dat2") %in% names(d)))
  expect_s3_class(d$dat1, "data.frame")
  expect_s3_class(d$dat2, "data.frame")
  expect_equal(nrow(d$dat1), 80)
  expect_true("trt" %in% names(d$dat1))
})

# ---------------------------------------------------------------------------
test_that("plot.pated() draws without error", {
  set.seed(17)
  n <- 200
  dat <- data.frame(
    pid = paste0("p", seq_len(n)),
    arm = rbinom(n, 1, 0.5),
    z   = rnorm(n)
  )
  dat$y <- 0.3 * dat$arm + 0.5 * dat$z + rnorm(n)
  fit <- pated(
    glm_(y ~ arm, family = "gaussian"),
    glm_(z ~ arm, family = "gaussian"),
    data = dat
  )
  pdf(file = tempfile(fileext = ".pdf"))
  on.exit(dev.off())
  expect_invisible(plot(fit))
})

# ---------------------------------------------------------------------------
# Bootstrap path: exercises LogRankAdapter$refit() and MMRMAdapter$refit(),
# neither of which is hit by any existing test. Use small nboot for speed.
test_that("bootstrap-mode coxph_ refits cleanly", {
  skip_if_not_installed("survival")
  library(survival)
  set.seed(19)
  n <- 200
  dat <- data.frame(
    pid   = paste0("p", seq_len(n)),
    arm   = rbinom(n, 1, 0.5),
    z     = rnorm(n),
    time  = rexp(n, rate = 0.2),
    event = rbinom(n, 1, 0.8)
  )
  fit <- jointCovariance(
    coxph_(Surv(time, event) ~ arm),
    glm_(z ~ arm, family = "gaussian"),
    data  = dat,
    nboot = 10,
    seed  = 1
  )
  expect_s3_class(fit, "jointCovariance")
  expect_true(all(is.finite(coef(fit))))
})

test_that("make_adapter() rejects unsupported spec class", {
  bogus <- structure(list(engine = "unknown", data_index = 1, id_col = "pid"),
                     class = c("jc_spec_bogus", "jc_spec"))
  expect_error(
    multipleOutcomes:::make_adapter(bogus, data_list = list(data.frame(pid = "x"))),
    "Unsupported spec type"
  )
})

test_that("bootstrap-mode logrank_ refits cleanly", {
  skip_if_not_installed("survival")
  library(survival)
  set.seed(21)
  n <- 200
  dat <- data.frame(
    pid   = paste0("p", seq_len(n)),
    arm   = rbinom(n, 1, 0.5),
    z     = rnorm(n),
    time  = rexp(n, rate = 0.2),
    event = rbinom(n, 1, 0.8)
  )
  fit <- jointCovariance(
    logrank_(Surv(time, event) ~ arm),
    glm_(z ~ arm, family = "gaussian"),
    data  = dat,
    nboot = 10,
    seed  = 1
  )
  expect_s3_class(fit, "jointCovariance")
  expect_true(all(is.finite(coef(fit))))
})

test_that("bootstrap-mode mmrm_ refits cleanly", {
  skip_if_not_installed("mmrm")
  long <- sim_repeated(n_subj = 80, n_visits = 3, seed = 22)
  fit <- jointCovariance(
    mmrm_(y ~ arm + us(visit | pid), reml = TRUE),
    data  = long,
    nboot = 5,
    seed  = 1
  )
  expect_s3_class(fit, "jointCovariance")
  expect_true(all(is.finite(coef(fit))))
})

# ---------------------------------------------------------------------------
# MMRMAdapter early-validation paths in mmrm_().
test_that("mmrm_() rejects bad control argument", {
  skip_if_not_installed("mmrm")
  expect_error(
    mmrm_(y ~ arm + us(visit | pid), control = "not a control"),
    "control must be an object of class mmrm_control"
  )
})

test_that("mmrm_() rejects Kenward-Roger with reml = FALSE", {
  skip_if_not_installed("mmrm")
  ctrl <- mmrm::mmrm_control(method = "Kenward-Roger")
  expect_error(
    mmrm_(y ~ arm + us(visit | pid), reml = FALSE, control = ctrl),
    "Kenward-Roger only works for REML"
  )
})

# ---------------------------------------------------------------------------
# sampleWithReplacement(): the "repeated-measurements per pid" branch.
test_that("sampleWithReplacement handles repeated measurements per pid", {
  set.seed(31)
  wide <- data.frame(
    pid = paste0("p", seq_len(40)),
    arm = rbinom(40, 1, 0.5),
    y   = rnorm(40)
  )
  long <- data.frame(
    pid = rep(paste0("p", seq_len(40)), each = 3),
    arm = rep(rbinom(40, 1, 0.5), each = 3),
    visit = factor(rep(1:3, 40)),
    y     = rnorm(120)
  )
  # Path used by bootstrapJointCovariance() with mixed wide+long input.
  fit <- jointCovariance(
    glm_(y ~ arm, family = "gaussian", data_index = 1),
    mmrm_(y ~ arm + us(visit | pid), reml = TRUE, data_index = 2),
    data  = list(wide, long),
    nboot = 5,
    seed  = 1
  )
  expect_s3_class(fit, "jointCovariance")
})

# Asymptotic (nboot = 0) path used to dispatch to parseTreatmentVariableFromFormula
# which silently produced an empty arm vector once jointCovariance() started
# requiring jc_spec specs. Make sure the call-based parser is now reached and
# the standard non-KM PATED pipeline works end-to-end.

test_that("pated() runs asymptotically with jc_spec inputs", {
  set.seed(11)
  n <- 200
  dat <- data.frame(
    pid  = paste0("p", seq_len(n)),
    trt  = rbinom(n, 1, 0.5),
    z1   = rnorm(n),
    z2   = rbinom(n, 1, 0.4)
  )
  dat$y <- 0.4 * dat$trt + 0.6 * dat$z1 + rnorm(n)

  fit <- pated(
    glm_(y  ~ trt, family = "gaussian", data_index = 1),
    glm_(z1 ~ trt, family = "gaussian", data_index = 1),
    glm_(z2 ~ trt, family = "binomial", data_index = 1),
    data = list(dat)
  )

  expect_s3_class(fit, "pated")
  expect_true("PATED" %in% fit$method)
  pated_row <- fit[fit$method == "PATED", ]
  expect_equal(nrow(pated_row), 1)
  expect_true(is.finite(pated_row$estimate))
  expect_true(is.finite(pated_row$stderr))
  expect_lt(pated_row$pvalue, 0.05)  # 0.4 effect on n=200 should be detectable
})

# g(S(t)), a transformed estimate of survival probability.
# See https://documentation.sas.com/doc/en/pgmsascdc/9.4_3.4/statug/statug_lifetest_details08.htm
gFunction <- function(conf.type = c('log', 'log-log', 'plain', 'logit', 'arcsin')){

  conf.type <- match.arg(conf.type)

  switch(conf.type,
    'log'     = function(x) log(x),
    'log-log' = function(x) log(-log(x)),
    'plain'   = function(x) x,
    'logit'   = function(x) log(x / (1 - x)),
    'arcsin'  = function(x) asin(sqrt(x))
  )
}

# Extract per-call metadata (function name, outcome label, treatment-arm
# variable, and any engine-specific argument like km_'s conf_type) from the
# `...` arguments passed to pated(). Uses the call AST rather than string
# surgery on deparse() output so that formulas with nested parens (e.g.,
# `y ~ arm + us(visit | pid)` for mmrm_) are parsed correctly.
parseTreatmentVariableFromCall <- function(...){

  calls <- as.list(substitute(list(...)))[-1L]
  for(i in seq_along(calls)){
    if(!is.call(calls[[i]])){
      stop(str_glue('{calls[[i]]} is not a call. '))
    }
  }

  func <- character(length(calls))
  outcome <- character(length(calls))
  arm <- character(length(calls))
  arg <- rep(NA_character_, length(calls))
  n_terms <- rep(1L, length(calls))

  for (idx in seq_along(calls)) {
    cstr <- calls[[idx]]
    func[idx] <- as.character(cstr[[1]])

    # Locate the formula argument: try `formula = ...` first, otherwise the
    # first positional argument after the function name.
    fml <- cstr$formula
    if (is.null(fml)) {
      arg_names <- names(cstr)
      if (is.null(arg_names)) arg_names <- rep("", length(cstr))
      positional <- which(arg_names == "")
      positional <- positional[positional > 1]
      if (length(positional) == 0) {
        stop(str_glue('No formula argument found in call: {deparse1(cstr)}'))
      }
      fml <- cstr[[positional[1]]]
    }
    if (!is.call(fml) || !identical(fml[[1]], as.name("~"))) {
      stop(str_glue('Argument is not a formula in call: {deparse1(cstr)}'))
    }

    outcome[idx] <- trimws(deparse1(fml[[2]]))

    # The treatment variable is the first plain variable name appearing on the
    # RHS, skipping anything that's only a covariance-structure helper (us,
    # ar1, cs, ...). all.vars() walks the AST and returns variable names in
    # order of appearance, so the first element is the treatment variable in
    # every formula we currently support.
    rhs_vars <- all.vars(fml[[3]])
    if (length(rhs_vars) == 0) {
      stop(str_glue('No treatment variable found on RHS of formula in call: {deparse1(cstr)}'))
    }
    arm[idx] <- rhs_vars[1]

    if (func[idx] == 'km_') {
      if (is.null(cstr$conf_type)) {
        stop('conf_type is missing in km_')
      }
      arg[idx] <- as.character(cstr$conf_type)
    }

    if (func[idx] == 'quantile_') {
      if (is.null(cstr$probs)) {
        stop('probs is missing in quantile_')
      }
      arg[idx] <- deparse(cstr$probs)
      n_terms[idx] <- length(eval(cstr$probs))
    }
  }

  data.frame(func = func, outcome = outcome, arm = arm,
             arg = arg, n_terms = n_terms,
             stringsAsFactors = FALSE)
}

#' Generating Data for Simulation and Testing
#' @description
#' `simulateMoData` generates data for simulation and testing purposes.
#' 
#' @param n an integer for total sample size of a randomized control trial of 
#' two arms. 
#' @param hr hazard ratio of treatment.
#' @param seed random seed. By default `NULL` for no seed being specified.
#' 
#' @export
simulateMoData <- function(n = 500, hr = 0.8, seed = NULL){
  
  set.seed(seed)
  
  rep_sigma <- matrix(.6, 3, 3) # correlation of repeated measurements
  diag(rep_sigma) <- 1
  repm <- data.frame(rmvnorm(n, sigma = rep_sigma))
  colnames(repm) <- paste0('repm', 1:ncol(repm))
  pmax_repm <- apply(repm, 1, max)
  
  cont <- rnorm(n) # continuous outcome
  bin <- rbinom(n, 1, .5) # binary outcome
  err <- rnorm(n, sd = .01)
  trt <- rbinom(n, 1, .5) # fully randomized arm assignment
  tte <- -log(runif(n)) / .05 / exp(log(hr) * trt + 5 * (.8 * cont + 2.5 * bin + 0 * pmax_repm) + err)
  tte <- tte^.35
  event <- rbinom(n, 1, .9) # event or censored
  tte <- tte * ifelse(event == 1, 1, runif(n))
  
  follow_up_time <- 100
  event[tte > follow_up_time] <- 0
  tte <- pmin(tte, follow_up_time)
  id <- paste0('uid-', 1:n)
  dat1 <- data.frame(cont, bin, tte, event, trt, id)
  dat2 <- data.frame(repm, trt, id) %>% 
    pivot_longer(
      cols = all_of(paste0('repm', 1:3)), 
      names_to = 'time', 
      values_to = 'repm') %>% 
    dplyr::select(.data$repm, .data$time, .data$trt, .data$id) %>% 
    as.data.frame()
  
  invisible(list(dat1 = dat1, dat2 = dat2))
  
}

make_adapter <- function(spec, data_list){
  
  stopifnot(inherits(spec, "jc_spec"))
  dat <- data_list[[spec$data_index]]
  
  if(!is.data.frame(dat)){
    stop("data[[data_index]] must be a data.frame")
  }
  
  if(inherits(spec, "jc_spec_glm")){
    return(GLMAdapter$new(spec, dat))
  }
  
  if(inherits(spec, "jc_spec_coxph")){
    return(CoxphAdapter$new(spec, dat))
  }
  
  if(inherits(spec, "jc_spec_logrank")){
    return(LogRankAdapter$new(spec, dat))
  }
  
  if(inherits(spec, "jc_spec_gee")){
    return(GEEAdapter$new(spec, dat))
  }
  
  if(inherits(spec, "jc_spec_mmrm")){
    return(MMRMAdapter$new(spec, dat))
  }
  
  if(inherits(spec, "jc_spec_km")){
    return(KMAdapter$new(spec, dat))
  }

  if(inherits(spec, "jc_spec_quantile")){
    return(QuantileAdapter$new(spec, dat))
  }

  stop("Unsupported spec type: ", paste(class(spec), collapse = ", "))
}

if(0){
library(mvtnorm)
library(data.table)
library(sandwich)
library(lmtest)

robust <- function(model){
  coeftest(model, vcov = vcovHC(model, type = "HC1"))
}


power <- function(estimate, se, level = .05){
  mean(pchisq((estimate / se)^2, df = 1, lower.tail = FALSE) < level)
}

n <- 500

adj1 <- NULL
adj2 <- NULL
adj3 <- NULL
se1 <- NULL
se2 <- NULL
se3 <- NULL
pat <- NULL
se0 <- NULL
for(i in 1:2000){
  x1 <- rbinom(n, 1, .5)
  x <- rmvnorm(n, sigma = matrix(c(2, 0, 0, 2), 2))
  x2 <- x[, 1]
  x3 <- x[, 2]
  x4 <- rnorm(n, sd = 2)
  y <- rbinom(n, 1, plogis(x1 * 1 + 1*x2 - 1*x3 + 3*log(abs(x2)) + x4))
  fit3 <- glm(y ~ x1 + x2 + x3 + x2^2, family = 'binomial')
  fit2 <- glm(y ~ x1 + x2, family = 'binomial')
  fit1 <- glm(y ~ x1, family = 'binomial')
  pfit <- pated(y ~ x1, x2 ~ x1, x3 ~ x1, 
                x2^2 ~ x1, #x3^2 ~ x1, 
                family = c('binomial', rep('gaussian', 3)), 
                data = data.frame(y,x1,x2,x3))
  adj1 <- c(adj1, coef(fit1)['x1'])
  adj2 <- c(adj2, coef(fit2)['x1'])
  adj3 <- c(adj3, coef(fit3)['x1'])
  pat <- c(pat, pfit$estimate[1])
  
  se1 <- c(se1, robust(fit1)[,2]['x1'])
  se2 <- c(se2, robust(fit2)[,2]['x1'])
  se3 <- c(se3, robust(fit3)[,2]['x1'])
  se0 <- c(se0, pfit$stderr[1])
}

#print('adj3')
mean(adj3)# %>% print()

#print('adj2')
mean(adj2)# %>% print()

#print('unadj')
mean(adj1)# %>% print()

#print('pated')
mean(pat)# %>% print()

power1 <- power(adj1, se1)
power2 <- power(adj2, se2)
power3 <- power(adj3, se3)
power0 <- power(pat, se0)

data.table(
  method = c('PATED', 'Adj1', 'Adj2', 'Adj3'),
  mean = c(mean(pat), mean(adj1), mean(adj2), mean(adj3)),
  'bias %' = ((c(mean(pat), mean(adj1), mean(adj2), mean(adj3))/mean(adj1) - 1)*100) %>% round(1),
  sd = c(sd(pat), sd(adj1), sd(adj2), sd(adj3)),
  se = c(mean(se0), mean(se1), mean(se2), mean(se3)),
  'power %' = round(100 * c(power0, power1, power2, power3), 1)
) %>% print



}

# Create curve of survival probability based on conventional KM method, or 
# PATED adjusted KM estimates.
# Refer to km_res or pated_res in `pated` about the format of `input`.
# Transformations are supported when computing confidence intervals.
# Refer to https://documentation.sas.com/doc/en/pgmsascdc/9.4_3.4/statug/statug_lifetest_details08.htm 
# or the LIFETEST proc in SAS for more details of transformation.
createKaplanMeierCurve <- function(input, title = '', transform = 'identity'){
  
  estimate <- input$estimate
  names(estimate) <- rownames(input)
  coordinate <- 
    data.frame(
      Time = extractKaplanMeierTimes(estimate, sort = FALSE),
      Strata = estimate %>% 
        names() %>%
        sapply(function(x) {sub('.*time_\\((.*)\\)_\\(.*\\)', '\\1', x)}),
      St = input$estimate,
      LCI = input$LCI,
      UCI = input$UCI
    )
  rownames(coordinate) <- NULL
  
  p <- ggplot(coordinate, aes(x = .data$Time, group = .data$Strata)) + 
    geom_ribbon(aes(ymin = .data$LCI, ymax = .data$UCI, fill = .data$Strata), alpha = .2) + 
    geom_step(aes(y = .data$St, color = .data$Strata)) + 
    scale_x_continuous(transform = transform) + 
    labs(x = str_glue('Time{ifelse(transform == "identity", "", paste(" (", transform, ")"))}'), y = 'Survival Probability S(t)', title = title) + 
    ylim(c(0, 1))
  
  attr(p, 'coordinate') <- coordinate
  p
  
}

#' Title Summarize an Analysis of Multiple Outcomes.
#' @description
#' Summarize an analysis of multiple outcomes. 
#' 
#' @param x an object returned by `jointCovariance()`.
#' @param ... for debugging only.
#'
#' @return an invisible object.
#' @export
#'
#' @examples
#' ## no example
print.summary.jointCovariance <- function(x, ...){
  
  cat('Call:\n')
  print(x$call)
  cat('\n')
  
  cat('Coefficients:\n')
  printCoefmat(x$coefficients, P.values = TRUE, has.Pvalue = TRUE)
  
}

#' 
BaseAdapter <- R6::R6Class(
  "BaseAdapter",
  public = list(
    
    type = NULL,
    spec = NULL, ## passed through ... in jointCovariance
    data = NULL,
    fit  = NULL,
    km_times = NULL, ## only for engine "km"
    
    # Standardized parameter vector ("beta") per engine
    estimate = NULL,
    score = NULL,
    inv_hess = NULL,
    n = NULL,
    sample_id = NULL,
    
    initialize = function(spec, data) {
      self$spec <- spec
      self$data <- data
    },
    
    fit_model = function(){
      stop("fit_model() not implemented.")
    },
    
    get_engine = function(){
      self$spec$engine
    },
    
    get_n = function(){
      if(is.null(self$n)){
        stop('sample size is not calculated. Debug it. ')
      }
      self$n
    },
    
    get_km_times = function(){
      self$km_times
    },
    
    get_sample_id = function(){
      if(is.null(self$sample_id)){
        stop('sample ID is missing. Debug it. ')
      }
      self$sample_id
    },
    
    get_coef = function(){
      if(is.null(self$estimate)) {
        stop("estimate is not found. Did you call $fit_model()?")
      }
      self$estimate
    },
    
    get_dim = function(){
      length(self$estimate)
    },
    
    get_coef_name = function(){
      names(self$estimate)
    },
    
    get_data_index = function(){
      self$spec$data_index
    },
    
    # Bootstrap refit: MUST return named vector aligned to coef_names
    refit = function(bdata){
      stop("refit() not implemented.")
    },
    
    # ---- stubs you will implement ----
    get_score = function(){

      if(self$type != 'logrank'){
        # Sanity check: at the MLE the per-subject score should average near 0.
        # GLMs converge with epsilon = 1e-8, which leaves a score-mean of up
        # to ~1e-5 on binomial / Poisson fits, so the threshold has to be
        # loose enough to accommodate that without flagging a true convergence.
        score_epsilon <- 1e-4
        stopifnot((colSums(self$score) / nrow(self$score)) %>% abs %>% max < score_epsilon)
      }
      self$score
    },
    
    get_bread = function(){
      self$inv_hess
    },
    
    get_type = function(){
      self$type
    },
    
    print = function(){
      print(self$type)
      print(self$estimate)
      print(head(self$score))
      print(self$inv_hess)
      invisible(self)
    }
  )
)

#' Extract Model Coefficients
#' @description
#' `coef` is a generic function. 
#' 
#' @param object an object returned by `jointCovariance()`.
#' @param model_index `NULL` if displaying coefficients of all fitted models;
#' otherwise, an integer indicating the fitted model.
#' @param ... for debugging only
#'
#' @return a vector of coefficient estimates
#' @export
#'
coef.jointCovariance <- function(object, model_index = NULL, ...){
  
  ret <- object$coefficients
  
  if(!is.null(model_index)){
    id_map <- object$id_map
    if(model_index > length(id_map)){
      stop('Invalid model index.')
    }
    
    ret <- ret[id_map[[model_index]]]
    names(ret) <- names(id_map[[model_index]])
  }
  
  ret
  
}

KMAdapter <- R6::R6Class(
  "KMAdapter",
  inherit = BaseAdapter,
  public = list(
    fit_model = function(){
      
      self$type <- 'km'
      
      self$fit <- fitKMCurve(self$spec$formula, self$data, self$spec$conf_type, self$spec$times)
      
      ## estimate
      self$n <- nrow(self$data)
      self$sample_id <- self$data[[self$spec$id_col]]
      self$km_times <- extractKaplanMeierTimes(self$fit)
      # build a plain named numeric so downstream c()/cov() see only the names
      # we care about and not the conf_type attribute carried on self$fit.
      est <- as.numeric(self$fit)
      names(est) <- names(self$fit)
      self$estimate <- est

      invisible(self)
    },
    
    refit = function(bdata){
      
      bfit <- fitKMCurve(self$spec$formula, bdata, self$spec$conf_type, 
                         times = self$get_km_times())
      
      bfit
    }
    
  )
)

#' ACTG 320 Clinical Trial Dataset
#' @description
#' actg dataset from Hosmer et al.
#' 
#' @name actg
#' @docType data
#' @format A data frame
#' \describe{
#' \item{id}{Identification Code}
#' \item{time}{Time to AIDS diagnosis or death (days).}
#' \item{censor}{Event indicator. 1 = AIDS defining diagnosis, 0 = Otherwise.}
#' \item{time_d}{Time to death (days)}
#' \item{censor_d}{Event indicator for death (only). 1 = Death, 0 = Otherwise.}
#' \item{tx}{Treatment indicator. 1 = Treatment includes IDV, 0 = Control group.}
#' \item{txgrp}{Treatment group indicator. 1 = ZDV + 3TC. 2 = ZDV + 3TC + IDV. 3 = d4T + 3TC. 4 = d4T + 3TC + IDV.}
#' \item{strat2}{CD4 stratum at screening. 0 = CD4 <= 50. 1 = CD4 > 50.}
#' \item{sex}{0 = Male. 1 = Female.}
#' \item{raceth}{Race/Ethnicity. 1 = White Non-Hispanic. 2 = Black Non-Hispanic. 3 = Hispanic. 4 = Asian, Pacific Islander. 5 = American Indian, Alaskan Native. 6 = Other/unknown.}
#' \item{ivdrug}{IV drug use history. 1 = Never. 2 = Currently. 3 = Previously.}
#' \item{hemophil}{Hemophiliac. 1 = Yes. 0 = No.}
#' \item{karnof}{Karnofsky Performance Scale. 100 = Normal; no complaint no evidence of disease. 90 = Normal activity possible; minor signs/symptoms of disease. 80 = Normal activity with effort; some signs/symptoms of disease. 70 = Cares for self; normal activity/active work not possible.}
#' \item{cd4}{Baseline CD4 count (Cells/Milliliter).}
#' \item{priorzdv}{Months of prior ZDV use (months).}
#' \item{age}{Age at Enrollment (years).}
#' }
#' 
#' @source ftp://ftp.wiley.com/public/sci_tech_med/survival
#' 
#' @references Hosmer, D.W. and Lemeshow, S. and May, S. (2008) Applied Survival Analysis: Regression Modeling of Time to Event Data: Second Edition, John Wiley and Sons Inc., New York, NY
#' 
#' @examples
#' data(actg)
#' 
NULL

#' @importFrom ggplot2 ggplot aes geom_point geom_abline geom_ribbon geom_step
#' @importFrom ggplot2 scale_x_continuous labs theme_minimal ylim
#' @importFrom mvtnorm rmvnorm
#' @importFrom tidyr pivot_longer
#' @importFrom survival survfit strata
#' @importFrom dplyr "%>%" select mutate
#' @importFrom tidyselect all_of
#' @importFrom rlang .data
#' @importFrom stringr str_glue
#' @importFrom stats model.matrix resid
#' @importFrom sandwich estfun bread
NULL

# 
# library(dplyr)
# library(momentfit)
# library(mvtnorm)
# set.seed(NULL)
# res <- NULL
# for(i in 1:1){
# n <- 5000
# z <- rbinom(n, 1, .5) # random assignment
# x <- rmvnorm(n, sigma = matrix(c(1, .5, .5, 1), 2))
# a <- rbinom(n, 1, plogis(0 + 3 * (2*z-1) + .6 * x[,1]))
# y <- 5*x[,1] + 1.5 * a + rnorm(n)
# data <- data.frame(y, z, x=x[,1], a)
# table(a,z)
# mean(a)
# mean(z)
# 
# res <- rbind(res, 
# data.frame(
#   z=t.test(x[,1]~z)$p.value, a=t.test(x[,1]~a)$p.value
# ))
# }
# 
# apply(res, 2, function(pr) {mean(pr < .05)}) %>% print()
# 
# workingModel <- function(b, z, x, a){
#   
#   gam0 <- b[4]
#   gam1 <- b[5]
#   gam2 <- b[6]
#   lin <- gam0 + gam1 * x + gam2 * z
#   exp(a * lin) / (1 + exp(lin))
#   
# }
# 
# conditionalProbabilityOfZ <- function(b, z, x, a){
#   
#   workingModel(b, z = z, x = x, a = a) / (workingModel(b, z = 0, x = x, a = a) + workingModel(b, z = 1, x = x, a = a))
#   
# }
# 
# moment3 <- function(b, resid, z){
#   resid * z
# }
# moment4 <- function(b, z, x, a){
#   a - workingModel(b, z = z, x = x, a = 1)
# }
# 
# moment5 <- function(b, z, x, a){
#   (a - workingModel(b, z = z, x = x, a = 1)) * x
# }
# 
# moment6 <- function(b, z, x, a){
#   (a - workingModel(b, z = z, x = x, a = 1)) * z
# }
# 
# moment7 <- function(z){
#   z - 1/2
# }
# 
# moment8 <- function(z, x){
#   x * (z - 1/2)
# }
# 
# moments <- function(b, data){
#   
#   y <- data$y
#   x <- data$x
#   a <- data$a
#   
#   bet0 <- b[1]
#   bet1 <- b[2]
#   bet2 <- b[3]
#   resid <- y - (bet0 + bet1 * x + bet2 * a)
#   u1 <- resid
#   u2 <- resid * x
#   u3 <- 
#     
#       moment3(b, resid, z = 0) * conditionalProbabilityOfZ(b, z = 0, x = x, a = a) + 
#         moment3(b, resid, z = 1) * conditionalProbabilityOfZ(b, z = 1, x = x, a = a)
#     
#   
#   u4 <- 
#     
#       moment4(b, z = 0, x = x, a = a) * conditionalProbabilityOfZ(b, z = 0, x = x, a = a) + 
#         moment4(b, z = 1, x = x, a = a) * conditionalProbabilityOfZ(b, z = 1, x = x, a = a)
#     
#   
#   u5 <- 
#     
#       moment5(b, z = 0, x = x, a = a) * conditionalProbabilityOfZ(b, z = 0, x = x, a = a) + 
#         moment5(b, z = 1, x = x, a = a) * conditionalProbabilityOfZ(b, z = 1, x = x, a = a)
#     
#   
#   u6 <- 
#     
#       moment6(b, z = 0, x = x, a = a) * conditionalProbabilityOfZ(b, z = 0, x = x, a = a) + 
#         moment6(b, z = 1, x = x, a = a) * conditionalProbabilityOfZ(b, z = 1, x = x, a = a)
#     
#   
#   u7 <- 
#     
#       moment7(z = 0) * conditionalProbabilityOfZ(b, z = 0, x = x, a = a) + 
#         moment7(z = 1) * conditionalProbabilityOfZ(b, z = 1, x = x, a = a)
#     
#   
#   u8 <- 
#     
#       moment8(z = 0, x = x) * conditionalProbabilityOfZ(b, z = 0, x = x, a = a) + 
#         moment8(z = 1, x = x) * conditionalProbabilityOfZ(b, z = 1, x = x, a = a)
#     
#   
#   ret <- cbind(u1, u2, u3, u4, u5, u6, u7, u8)
#   
#   #print(b)
#   
#   sum(apply(ret, 2, mean)^2)
# }
# 
# #set.seed(2)
# #fit <- momentfit::gmm4(moments, x = data, theta0 = rnorm(6), type = 'twostep')
# 
# #set.seed(1)
# t0 <- rnorm(6)
# t00 <- t0
# 
# fn <- NULL
# while(TRUE){
#   print('eps: ')
#   print(eps)
#   print('t0: ')
#   print(t0)
#   print('fn')
#   fn <- c(fn, moments(t0, data))
#   print(tail(fn, 1))
#   print('================')
#   g <- numDeriv::grad(moments, t0, data = data)
#   t0 <- t0 - .5 * g
#   eps <- max(abs(.5 * g))
#   if(eps < 1e-4){
#     break
#   }
# }
# 
# 
# moments(t1, data)

#' Creating Objects of Logrank Test
#' 
#' @description
#' `logrank_` is a wrapper function `survival::coxph` to create an object to be 
#' passed into `jointCovariance`, the main function of this package through its 
#' argument `...`. Logrank test is the score test under the proportional hazards 
#' regression model. The object defines how a logrank test would be computed. 
#' 
#' @param formula see `formula` in `survival::coxph`.
#' @param ties character string specifying the method for tie handling. One of
#' `"efron"` (default), `"breslow"`, or `"exact"`. Passed through to
#' `survival::coxph`.
#' @param data_index integer. Index of the data frame in the `data` argument of
#' `jointCovariance` to be used when computing testing statistic of logrank 
#' test.
#' 
#' @details
#' 
#' Not all arguments of `survival::coxph` are supported in `logrank_` due to the 
#' complexity in handling environment and scope, which is particularly difficult 
#' for arguments like `weights`, `subset`, etc. 
#' 
#' @export
logrank_ <- function(formula, ties = c('efron', 'breslow', 'exact'), data_index = 1){
  
  ties <- match.arg(ties)
  data_index <- validate_data_index(data_index)

  structure(
    list(
      engine = "longrank",
      formula = formula,
      ties = ties,
      data_index = data_index,
      id_col = 'pid'
    ),
    class = c("jc_spec_logrank", "jc_spec")
  )
}

#' Object Summaries
#' @description
#' `summary` method for class `jointCovariance`.
#' 
#' @param object an object returned by `jointCovariance()`.
#' @param model_index `NULL` if displaying summary of all fitted models; otherwise,
#' an integer indicating the fitted model.
#' @param ... for debugging only
#'
#' @return a list
#' @export
#' 
summary.jointCovariance <- function(object, model_index = NULL, ...){
  
  if(is.null(object$mcov)){
    se <- apply(object$bootstrap_estimate, 2, sd, na.rm = TRUE)
  }else{
    se <- sqrt(diag(object$mcov))
  }
  
  tval <- coef(object) / se
  
  TAB <- cbind(Estimate = coef(object),
               StdErr = se,
               t.value = tval,
               p.value = pchisq(tval^2, df = 1, lower.tail = FALSE))
  colnames(TAB) <- c('Estimate', 'Std. Error', 'z value', 'Pr(>|z|)')
  
  id_map <- object$id_map
  if(!is.null(model_index)){
    if(model_index > length(id_map)){
      stop('Invalid model index.')
    }
    TAB <- TAB[id_map[[model_index]], , drop = FALSE]
    rownames(TAB) <- names(id_map[[model_index]])
  }else{
    rownames(TAB) <-
      lapply(
        1:length(id_map),
        function(mid){
          paste0('model_', mid, '_', names(id_map[[mid]]))
        }
      ) %>% unlist()
  }
  
  res <- list(call = object$call,
              coefficients = TAB)
  
  class(res) <- 'summary.jointCovariance'
  res
  
}

#' Creating Objects of Generalized Linear Models
#' 
#' @description
#' `glm_` is a wrapper function of `stats::glm` to create an object to be passed 
#' into `jointCovariance`, the main function of this package through its 
#' argument `...`. The object defines how a GLM model would be fitted. 
#' 
#' @param formula see `formula` in `stats::glm`. 
#' @param family currently supports `"gaussian"` or `"binomial"`. Other families 
#' are under testing. 
#' @param data_index integer. Index of the data frame in the `data` argument of
#' `jointCovariance` to be used when fitting a generalized linear model. 
#' 
#' @details
#' 
#' Not all arguments of `stats::glm` are supported in `glm_` due to the 
#' complexity in handling environment and scope, which is particularly difficult 
#' for arguments like `weights`, `subset`, etc. 
#' 
#' @export
glm_ <- function(formula, family, data_index = 1) {
  
  if(!inherits(formula, 'formula')){
    stop('"formula" in glm_ must be a R formula. ')
  }
  data_index <- validate_data_index(data_index)

  structure(
    list(
      engine = "glm",
      formula = formula,
      outcome = trimws(sub("\\s*~.*", "", deparse1(formula))),
      family  = family,
      data_index = data_index,
      id_col = 'pid'
    ),
    class = c("jc_spec_glm", "jc_spec")
  )
}

#' Creating Objects of Kaplan-Meier Curve
#' 
#' @description
#' `km_` is a wrapper function creating an object of Kaplan-Meier curve to be 
#' passed into `jointCovariance`, the main function of this package through its 
#' argument `...`. The object defines how a Kaplan-Meier curve would be fitted. 
#' 
#' @param formula a formula created by `survival::Surv()`. 
#' @param times numeric vector of time. Survival probabilities at `times` are 
#' computed (with g-transformation defined by `conf_type`). 
#' @param conf_type character. Type of confidence interval. It must be one of 
#' `"log"`, `"log-log"`, `"plain"`, `"logit"`, or `"arcsin"`. 
#' @param data_index integer. Index of the data frame in the `data` argument of
#' `jointCovariance` to be used when fitting a generalized linear model. 
#' 
#' @details
#' 
#' Usually, g-transformation is applied to the survival probability `S(t)` 
#' to obtain pointwise confidence interval of a Kaplan-Meier curve. This 
#' can be achieved by specifying `conf_type`. For identity transformation, 
#' use `conf_type = "plain"`. 
#' 
#' This function can only be used with `jointCovariance` when the bootstrap 
#' method is used to estimate variance-covariance matrix of multiple outcome 
#' models.  
#' 
#' @export
km_ <- function(formula, times = NULL, conf_type, data_index = 1){
  
  if(!(conf_type %in% c('log','log-log','plain', 'logit', 'arcsin'))){
    stop("conf.type takes value from 'log','log-log','plain', 'logit', 'arcsin'")
  }
  data_index <- validate_data_index(data_index)

  structure(
    list(
      engine = "km",
      formula = formula, 
      times = times, 
      conf_type = conf_type,
      data_index = data_index,
      id_col = 'pid'
    ),
    class = c("jc_spec_km", "jc_spec")
  )
}

# df is of length that equals to the number of specified models
# df[i] is the number of parameters in model i
# This function returns index of parameters of model i in the vector of 
# parameters of all models
IDMapping <- function(df, i, var_name = NULL){
  if(i == 1){
    ids <- 1:df[1]
  }else{
    ids <- (sum(df[1:(i - 1)]) + 1):sum(df[1:i])
  }
  
  if(!is.null(var_name) && length(var_name) == length(ids)){
    names(ids) <- var_name
  }
  
  ids
}

QuantileAdapter <- R6::R6Class(
  "QuantileAdapter",
  inherit = BaseAdapter,
  public = list(

    fit_model = function() {

      self$type <- "quantile"
      self$estimate <- compute_quantile_diff(self$spec$formula,
                                             self$data,
                                             self$spec$probs)

      self$n <- nrow(self$data)
      self$sample_id <- self$data[[self$spec$id_col]]

      invisible(self)
    },

    refit = function(bdata) {
      compute_quantile_diff(self$spec$formula, bdata, self$spec$probs)
    }
  )
)

# Two-group quantile difference. Returns a named vector with one entry per
# probability, where the names follow the legacy `arm_25%`, `arm_50%`, ...
# convention so pated()'s arm-lookup regex picks them up.
compute_quantile_diff <- function(formula, data, probs) {

  mf <- stats::model.frame(formula, data)
  outcome <- mf[[1L]]
  group   <- mf[[2L]]
  if (length(unique(group)) != 2L) {
    stop("Only two groups are allowed in quantile_.")
  }

  group_name <- names(mf)[2L]
  split_data <- split(outcome, group)

  qs <- do.call(
    cbind,
    lapply(split_data,
           function(sdata) stats::quantile(sdata, probs = probs))
  )

  qdiff <- qs[, 1] - qs[, 2]
  names(qdiff) <- paste0(group_name, "_", rownames(qs))
  qdiff
}

#' Creating Objects of Mixed Models for Repeated Measures
#' 
#' @description
#' `mmrm_` is a wrapper function of `mmrm::mmrm` to create an object to be passed 
#' into `jointCovariance`, the main function of this package through its 
#' argument `...`. The object defines how a MMRM model would be fitted. 
#' 
#' 
#' @param formula see `formula` in `mmrm::mmrm`.
#' @param covariance see `covariance` in `mmrm::mmrm`. 
#' @param reml see `reml` in `mmrm::mmrm`. 
#' @param control see `control` in `mmrm::mmrm`. 
#' @param ... see `...` in `mmrm::mmrm`.
#' @param data_index integer. Index of the data frame in the `data` argument of
#' `jointCovariance` to be used when fitting a GEE model. 
#' 
#' @details
#' 
#' The argument `weights` of `mmrm::mmrm` is supported in `mmrm_` due to the 
#' complexity in handling environment and scope. 
#' 
#' Please always refer to help document of `mmrm::mmrm` before using `mmrm_`. 
#' For example, time variable and observation ID must be factor variables in 
#' some cases, otherwise error may be prompted. Users can call `mmrm::mmrm` 
#' using the same arguments being passed to `mmrm_` to check validity. 
#' 
#' @export
mmrm_ <- function(formula, covariance = NULL, reml = TRUE, 
                  control = mmrm::mmrm_control(...), ..., 
                  data_index = 1){
  
  if(!missing(control) && length(list(...)) > 0){
    stop('Fine-grained control can be specified either via ... or via control, 
         not both. See the Details section in ?mmrm. ')
  }
  
  if(!inherits(control, 'mmrm_control')){
    stop('control must be an object of class mmrm_control 
         (created by mmrm::mmrm_control()). ', 
         call. = FALSE)
  }
  
  if(!is.list(control$optimizers) || length(control$optimizers) < 1L){
    stop("control$optimizes must be a non-empty list (as produced by mmrm::mmrm_control()).",
         call. = FALSE)
  }
  
  if(control$method %in% c("Kenward-Roger", "Kenward-Roger-Linear") && !isTRUE(reml)){
    stop("Kenward-Roger only works for REML.", call. = FALSE)
  }
  
  ## otherwise mmrm::component(object, 'score_per_subject') doesn't return scores
  if(!identical(control$method, 'Residual')){
    control$method <- 'Residual'
    message('method in mmrm_control is re-set to be "Residual"')
  }
  
  if(!identical(control$vcov, 'Empirical')){
    control$vcov <- 'Empirical'
    message('vcov in mmrm_control is re-set to be "Empirical"')
  }

  data_index <- validate_data_index(data_index)

  structure(
    list(
      engine = "mmrm",
      formula = formula,
      covariance = covariance, 
      reml = reml,
      control = control,
      data_index = data_index,
      id_col = 'pid'
    ),
    class = c("jc_spec_mmrm", "jc_spec")
  )
}

#' Creating Objects of Generalized Estimation Equation Model
#' 
#' @description
#' `gee_` is a wrapper function of `gee::gee` to create an object to be passed 
#' into `jointCovariance`, the main function of this package through its 
#' argument `...`. The object defines how a GEE model would be fitted. 
#' 
#' This package does not import the package `gee`. Instead, codes of `gee` are 
#' modified and integrated to compute score and information matrix. Thus, users 
#' does not need to install the package `gee` to use this package.  
#' 
#' @param formula see `formula` in `gee::gee`. 
#' @param family see `family` in `gee::gee`. 
#' @param corstr see `corstr` in `gee::gee`. 
#' @param R see `R` in `gee::gee`. 
#' @param b see `b` in `gee::gee`. 
#' @param Mv see `Mv` in `gee::gee`. 
#' @param data_index integer. Index of the data frame in the `data` argument of
#' `jointCovariance` to be used when fitting a GEE model. 
#' 
#' @details
#' 
#' Not all arguments of `stats::gee` are supported in `gee_` due to the 
#' complexity in handling environment and scope, which is particularly difficult 
#' for arguments like `subset`, etc. 
#' 
#' @export
gee_ <- function(formula, family, corstr, R = NULL, b = NULL, Mv = 1, data_index = 1) {
  data_index <- validate_data_index(data_index)
  structure(
    list(
      engine = "gee",
      formula = formula,
      family  = family,
      corstr = corstr, 
      R = R, 
      b = b,
      Mv = Mv,
      data_index = data_index,
      id_col = 'pid'
    ),
    class = c("jc_spec_gee", "jc_spec")
  )
}

# Cover the dispatch branches in src/ugee.c that handle different
# (family, corstr) combinations. The default fast gee_ smoke test exercises
# only (gaussian, exchangeable); these variants hit the other arms of the
# C-level switch on family and correlation structure.
#
# Each fit is a smoke test: we only assert the engine runs end-to-end through
# jointCovariance() and produces a finite coefficient. No PATED here, so we
# don't have to worry about arm-independence of prognostics.

# Build a longitudinal dataset with the appropriate outcome type for the
# requested family. Within-subject correlation comes from a shared per-subject
# random effect plus AR(1) noise.
make_long_for_family <- function(family, n_subj = 80, n_visits = 3,
                                 beta_arm = 0.30, seed = 1) {
  set.seed(seed)
  pid   <- paste0("p", seq_len(n_subj))
  arm   <- rbinom(n_subj, 1, 0.5)
  b     <- rnorm(n_subj, sd = 0.5)            # subject random intercept
  eta   <- outer(beta_arm * arm + b, rep(1, n_visits)) +
           matrix(rnorm(n_subj * n_visits, sd = 0.3), n_subj, n_visits)

  y <- switch(family,
    gaussian = as.vector(t(eta)),
    binomial = as.integer(as.vector(t(plogis(eta))) > runif(n_subj * n_visits)),
    poisson  = stats::rpois(n_subj * n_visits, lambda = exp(as.vector(t(eta))))
  )

  data.frame(
    pid   = rep(pid, each = n_visits),
    arm   = rep(arm, each = n_visits),
    visit = factor(rep(seq_len(n_visits), times = n_subj)),
    y     = y
  )
}

# (family, corstr, Mv) tuples to exercise. Each combination touches different
# code in src/ugee.c's main loop.
gee_configs <- list(
  list(family = "gaussian", corstr = "independence"),
  list(family = "gaussian", corstr = "AR-M",          Mv = 1),
  list(family = "gaussian", corstr = "unstructured"),
  list(family = "gaussian", corstr = "stat_M_dep",    Mv = 1),
  list(family = "binomial", corstr = "exchangeable"),
  list(family = "binomial", corstr = "independence"),
  list(family = "poisson",  corstr = "exchangeable"),
  list(family = "poisson",  corstr = "independence")
)

for (cfg in gee_configs) {
  label <- sprintf("gee_(%s, %s)", cfg$family, cfg$corstr)
  test_that(paste(label, "fits via jointCovariance"), {
    dat <- make_long_for_family(cfg$family, seed = 100L + which(
      sapply(gee_configs, identical, cfg)
    ))
    args <- c(list(formula = y ~ arm,
                   family  = cfg$family,
                   corstr  = cfg$corstr),
              if (!is.null(cfg$Mv)) list(Mv = cfg$Mv))
    spec <- do.call(gee_, args)
    fit  <- jointCovariance(spec, data = dat)
    expect_s3_class(fit, "jointCovariance")
    expect_true(all(is.finite(coef(fit))))
    V <- vcov(fit)
    expect_true(is.matrix(V))
    expect_lt(max(abs(V - t(V))), 1e-6)
  })
}

# for HR in coxph, use mean = TRUE
# for logrank, use mean = FALSE
imatCoxph <- function(model, mean = TRUE){
  imat <- survival::coxph.detail(model)$imat
  n <- nrow(matrix(resid(model, type = 'score')))
  if('numeric' %in% class(imat)){
    imat <- sum(imat) / n
  }else{
    n <- nrow(resid(model, type = 'score'))
    imat <- rowSums(imat, dims = 2) / n
  }
  if(!mean){
    imat <- imat * n
  }
  
  imat
}

GLMAdapter <- R6::R6Class(
  "GLMAdapter",
  inherit = BaseAdapter,
  public = list(
    fit_model = function(){
      
      self$type <- 'glm'
      
      self$fit <- stats::glm(formula = self$spec$formula, 
                             family = self$spec$family, 
                             data = self$data)
      
      ## estimate
      self$estimate <- stats::coef(self$fit)
      
      ## score
      self$score <- sandwich::estfun(self$fit)
      self$n <- stats::nobs(self$fit)
      stopifnot(self$n == nrow(self$score))
      
      kept_ids <- setdiff(seq_len(nrow(self$data)), self$fit$na.action)
      rownames(self$score) <- self$data[[self$spec$id_col]][kept_ids]
      self$sample_id <- rownames(self$score)
      
      ## inv_hess
      self$inv_hess <- -sandwich::bread(self$fit)
      
      invisible(self)
    },
    
    refit = function(bdata){
      bfit <- stats::glm(formula = self$spec$formula, 
                         family = self$spec$family, 
                         data = bdata)
      stats::coef(bfit)
    }
    
  )
)

GEEAdapter <- R6::R6Class(
  "GEEAdapter",
  inherit = BaseAdapter,
  public = list(
    fit_model = function(){
      
      self$type <- 'gee'
      
      id <- as.factor(self$data[[self$spec$id_col]])
      dat <- self$data
      dat$id <- id
      
      suppressMessages(
        capture.output(
          self$fit <- gee(formula = self$spec$formula, 
                          id = id, 
                          data = dat, 
                          family = self$spec$family, 
                          corstr = self$spec$corstr,
                          R = self$spec$R, 
                          b = self$spec$b, 
                          Mv = self$spec$Mv), 
          file = NULL
        )
      )
      
      ## estimate
      self$estimate <- stats::coef(self$fit)
      
      ## score
      self$score <- self$fit$score
      
      rownames(self$score) <- self$fit$score_id
      self$sample_id <- rownames(self$score)
      self$n <- nrow(self$score)
      
      ## inv_hess
      self$inv_hess <- solve(-self$fit$hess / self$n)
      
      invisible(self)
    },
    
    refit = function(bdata){
      
      ## multipleOutcomes::gee() is modified from gee::gee(), where id 
      ## is very tricky. It actually want it be numeric or factor, but has never 
      ## been mentioned in its manual. To see this, search code line: 
      ## id <- as.double(id)
      ## in R/gee.R in this package or R/ugee.R in the gee package. 
      ## That line of code could be in trouble if id is a character vector. 
      id <- as.factor(bdata[[self$spec$id_col]])
      bdata$id <- id
      
      suppressMessages(
        capture.output(
          bfit <- gee(formula = self$spec$formula, 
                      id = id, 
                      data = bdata, 
                      family = self$spec$family, 
                      corstr = self$spec$corstr,
                      R = self$spec$R, 
                      b = self$spec$b, 
                      Mv = self$spec$Mv), 
          file = NULL
        )
      )
      
      stats::coef(bfit)
    }
    
  )
)

#' Creating Objects of Proportional Hazards Regression Model
#' 
#' @description
#' `coxph_` is a wrapper function of `survival::coxph` to create an object to be 
#' passed into `jointCovariance`, the main function of this package through its 
#' argument `...`. The object defines how a proportional hazard model would be 
#' fitted. 
#' 
#' @param formula see `formula` in `survival::coxph`. 
#' @param data_index integer. Index of the data frame in the `data` argument of
#' `jointCovariance` to be used when fitting a proportional hazards model. 
#' 
#' @details
#' 
#' Not all arguments of `survival::coxph` are supported in `coxph_` due to the 
#' complexity in handling environment and scope, which is particularly difficult 
#' for arguments like `weights`, `subset`, etc. 
#' 
#' @export
coxph_ <- function(formula, data_index = 1) {
  data_index <- validate_data_index(data_index)
  structure(
    list(
      engine = "coxph",
      formula = formula,
      data_index = data_index,
      id_col = 'pid'
    ),
    class = c("jc_spec_coxph", "jc_spec")
  )
}

# Reproducible piecewise-exponential dataset with a delayed treatment effect,
# matching the legacy regression fixture saved under inst/testdata/.
gen_delayed_effect_data <- function(n = 600,
                                    lam0       = 0.20,
                                    tau_delay  = 3,
                                    HR_late    = 0.55,
                                    admin_tau  = 12,
                                    gamma1     = -0.40,
                                    gamma2     = -0.50,
                                    seed       = 20260516,
                                    id_col     = "pid") {
  set.seed(seed)

  arm <- rbinom(n, 1, 0.5)
  x_raw <- matrix(rnorm(2 * n), ncol = 2) %*% chol(matrix(c(1, .4, .4, 1), 2))
  x1 <- as.numeric(x_raw[, 1])
  x2 <- as.numeric(x_raw[, 2] > 0)
  z1 <- rnorm(n)

  lam_subj <- lam0 * exp(gamma1 * x1 + gamma2 * x2)
  u <- runif(n)
  S_at_delay <- exp(-lam_subj * tau_delay)

  surv_time <- numeric(n)
  early <- u > S_at_delay
  surv_time[early] <- -log(u[early]) / lam_subj[early]
  late <- !early
  eta <- ifelse(arm[late] == 1, HR_late, 1)
  surv_time[late] <- tau_delay +
    (-log(u[late]) - lam_subj[late] * tau_delay) / (lam_subj[late] * eta)

  event <- as.integer(surv_time <= admin_tau)
  time  <- pmin(surv_time, admin_tau)

  out <- data.frame(
    time  = time,
    event = event,
    arm   = arm,
    x1    = x1,
    x2    = x2,
    z1    = z1
  )
  out[[id_col]] <- paste0("subj-", seq_len(n))
  out
}

bootstrapJointCovariance <- function(..., data, nboot, compute_cov, seed){
  
  specs <- list(...)
  if(length(specs) == 0){
    stop("No model specs provided in ...")
  }
  
  if(!is.list(data)){
    stop("data must be a list of data.frames.")
  }
  
  adapters <- lapply(specs, make_adapter, data_list = data)
  adapters <- lapply(adapters, function(ad){ ad$fit_model(); ad })
  
  for(ad in adapters){
    if(ad$get_engine() %in% 'km'){
      if(compute_cov){
        compute_cov <- FALSE
        message('compute_cov is set to FALSE as km_ is in use. ', 
                'You can compute the covariance matrix using $bootstrap_estimate. ')
      }
      break
    }
  }
  
  n_models <- length(adapters)
  bet <- NULL
  df <- NULL
  vars <- vector('list', n_models)
  id_map <- NULL
  estimate <- NULL
  
  set.seed(seed)
  for(b in 0:nboot){
    if(b == 0){
      bdata <- data
    }else{
      bdata <- sampleWithReplacement(data)
    }
    
    model_fit <- NULL
    for (i in 1:n_models){
      data_index <- adapters[[i]]$get_data_index()
      fit <- adapters[[i]]$refit(bdata[[data_index]])
      names(fit) <- paste0('model_', i, '_', names(fit))
      model_fit <- c(model_fit, fit)
      
      if(b == 0){
        df[i] <- length(fit)
        id_map[[i]] <- IDMapping(df, i, adapters[[i]]$get_coef_name())
      }
    }
    
    if(b == 0){
      bet <- model_fit
    }else{
      estimate <- rbind(estimate, model_fit)
    }
    
  }
  
  rownames(estimate) <- NULL
  
  if(compute_cov){
    mcov <- cov(estimate)
    rownames(mcov) <- NULL
    colnames(mcov) <- NULL
    bootstrap_estimate <- NULL
  }else{
    mcov <- NULL
    bootstrap_estimate <- estimate
  }
  
  n_shared_sample_sizes <- matrix(0, n_models, n_models)
  
  for(i in 1:n_models){
    idx1 <- adapters[[i]]$get_data_index()
    n_shared_sample_sizes[i, i] <- adapters[[i]]$get_n()
    for(j in i:n_models){
      idx2 <- adapters[[j]]$get_data_index()
      n_shared_sample_sizes[i, j] <- intersect(adapters[[i]]$get_sample_id(), adapters[[j]]$get_sample_id()) %>% length()
      n_shared_sample_sizes[j, i] <- n_shared_sample_sizes[i, j]
    }
  }
  
  bet <- unname(bet)
  
  ret <-
    list(
      coefficients = bet,
      mcov = mcov,
      bootstrap_estimate = bootstrap_estimate,
      id_map = id_map,
      n_shared_sample_sizes = n_shared_sample_sizes
    )
  
  invisible(ret)
}

fitKMCurve <- function(formula, data, conf_type, times = NULL){

  tr <- try(fit <- survival::survfit(formula, data = data,
                                     type = 'kaplan-meier', se.fit = FALSE,
                                     conf.type = 'none'))
  if('try-error' %in% class(tr)){
    stop('Unable to fit K-M curve. ')
  }

  if(is.null(times)){
    times <- sort(unique(fit$time))
  }

  # extend = TRUE carries the last observed step value forward when a requested
  # time is past the last event in a stratum; this is the conventional KM
  # extrapolation and avoids NAs leaking into the bootstrap covariance matrix.
  pred <- summary(fit, times = times, extend = TRUE)
  if(is.null(pred$strata)){
    pred$strata <- 'trt=.'
  }
  pred$surv <- gFunction(conf_type)(pred$surv)

  trans_St <- pred$surv
  names(trans_St) <- paste0('time_(', pred$strata, ')_(', pred$time, ')')
  attr(trans_St, 'conf_type') <- conf_type

  trans_St
}

#' Fitting Regression Models for Multiple Outcomes and Returning the Matrix of
#' Covariance
#' @description
#' `jointCovariance` can fit different types of models for multiple outcomes 
#' simultaneously and return model parameters and variance-covariance matrix 
#' for further analysis. 
#'
#' @param ... objects returned by `glm_()`, `coxph_()`, `logrank_()`, `gee_()` 
#' and `gmm_()`. 
#' @param data a data frame if all models are fitted on the same dataset;
#' otherwise a list of data frames for fitting models in `...`. Note that a
#' dataset can be used to fit multiple models, thus, `length(data)` is unnecessary
#' to be equal to the number of models in `...`. The row names in a data frame
#' are not treated as subject IDs. Instead, all data frame should consist of 
#' a column `pid` as subject IDs. For any two records in different
#' data frames that correspond to the same subject, their values in `pid` should 
#' be consistent. For data frames to be used by GEE, `pid` defines clusters. 
#' See `id` and the Details section of `gee:gee`. 
#' @param nboot non-zero integer if bootstrap is adopted. By default 0.
#' @param compute_cov logic. If \code{TRUE} and `nboot > 0`, 
#' empirical covariance matrix is computed using bootstrap estimate and 
#' returned. Bootstrap estimate will be abandoned. If 
#' \code{FALSE}, bootstrap estimate will be returned and no empirical covariance 
#' matrix is computed. 
#' @param seed random seed when generate bootstrap data. 
#'
#' @return It returns an object of class "jointCovariance", which is a list
#' containing the following components:
#' \tabular{ll}{
#' \code{coefficients} \tab an unnamed vector of coefficients of all fitted models.
#' Use `id_map` for variable mapping. \cr
#' \tab \cr
#' \code{mcov} \tab a unnamed matrix of covariance of `coefficients`. Use `id_map`
#' for variable mapping. \cr
#' \tab \cr
#' \code{id_map} \tab a list mapping the elements in `coefficients` and `mcov` to
#' variable names. \cr
#' \tab \cr
#' \code{n_shared_sample_sizes} \tab a matrix of shared sample sizes between datasets
#' being used to fit the models. \cr
#' \tab \cr
#' \code{call} \tab the matched call.\cr
#' }
#' @export
#'
#' @examples
#' ## More examples can be found in the vignettes.
#' library(survival)
#' library(mvtnorm)
#' library(tidyr)
#' genData <- function(seed = NULL){
#'   
#'   set.seed(seed)
#'   n <- 300
#'   sigma <- matrix(.7, 4, 4)
#'   diag(sigma) <- 1
#'   v <- rmvnorm(n, sigma = sigma)
#'   x1 <- v[, 1]
#'   x2 <- v[, 2]
#'   z1 <- (v[, 3] > 0) + 0
#'   z2 <- v[, 4]
#'   
#'   trt <- rbinom(n, 1, .5)
#'   
#'   bet <- c(-.3,.3)
#'   y <- -log(runif(n))/
#'     exp(-.3 * x1 + .3 * x2 + z1 * .5 - z2 * .3 + .1 * trt + rnorm(n))
#'   
#'   z1[sample.int(n, 50)] <- NA
#'   z2[sample.int(n, 50)] <- NA
#'   x1[sample.int(n, 50)] <- NA
#'   x2[sample.int(n, 50)] <- NA
#'   death <- ifelse(y > 2, 0, 1)
#'   y[y > 2] <- 2
#'   
#'   pid <- paste0('pid-', 1:n)
#'   ret <- data.frame(
#'     y = y, trt = trt, 
#'     z1 = z1, z2 = z2, 
#'     x1 = x1, x2 = x2, 
#'     death, pid)
#'   ret
#' }
#' 
#' dat1 <- genData()
#' 
#' ## create a dataset with repeated measurements x
#' dat2 <- dat1 %>% pivot_longer(c(x1, x2), names_to='visit', values_to='x') %>% 
#'   dplyr::select(x, trt, visit, pid) %>% as.data.frame()
#' 
#' dat2$visit <- as.factor(dat2$visit)
#' dat2$pid <- as.factor(dat2$pid)
#' 
#' fit <- jointCovariance(
#'   coxph_(Surv(time = y, event = death) ~ trt, data_index = 1),
#'   logrank_(Surv(time = y, event=death) ~ trt, data_index = 1),
#'   glm_(z1 ~ trt, family = 'binomial', data_index = 1), 
#'   glm_(z2 ~ trt, family = 'gaussian', data_index = 1), 
#'   mmrm_(x ~ trt + us(visit | pid), reml = TRUE, data_index = 2), 
#'   gee_(x ~ trt, family = 'gaussian', corstr = 'independence', data_index = 2), 
#'   data = list(dat1, dat2))
#' 
#' fit
#'
#' \donttest{
#' bfit <-jointCovariance(
#'   coxph_(Surv(time=y, event=death) ~ trt, data_index = 1),
#'   logrank_(Surv(time=y, event=death) ~ trt, data_index = 1),
#'   glm_(z1 ~ trt, family = 'binomial', data_index = 1),
#'   glm_(z2 ~ trt, family = 'gaussian', data_index = 1),
#'   mmrm_(x ~ trt + us(visit | pid), reml = TRUE, data_index = 2),
#'   gee_(x ~ trt, family = 'gaussian', corstr = 'independence', data_index = 2),
#'   data = list(dat1, dat2), nboot = 10)
#'
#' summary(bfit)
#'
#' ## km_() and quantile_() require nboot > 0 because they have no
#' ## closed-form score. compute_cov is forced to FALSE for km_().
#' ## When all models share one dataset, `data_index` and `list(...)`
#' ## can be omitted.
#' kfit <- jointCovariance(
#'   km_(Surv(time = y, event = death) ~ trt, conf_type = 'log',
#'       times = c(0.5, 1, 1.5)),
#'   glm_(z1 ~ trt, family = 'binomial'),
#'   data = dat1, nboot = 30, seed = 1)
#'
#' qfit <- jointCovariance(
#'   quantile_(y ~ trt, probs = c(0.25, 0.5, 0.75)),
#'   glm_(z2 ~ trt, family = 'gaussian'),
#'   data = dat1, nboot = 30, seed = 1)
#' }
#'
jointCovariance <- function(..., data, nboot = 0, compute_cov = TRUE, seed = NULL){
    
    UseMethod('jointCovariance')
    
  }

#' @useDynLib multipleOutcomes, .registration = TRUE
#' @importFrom stats gaussian model.extract quantile
#' @importFrom utils capture.output
gee <- function(formula = formula(data), id = id, data = parent.frame(),
                subset, na.action, R = NULL, b = NULL, tol = 0.001, maxiter = 25,
                family = gaussian, corstr = "independence", Mv = 1,
                silent = TRUE, contrasts = NULL, scale.fix = FALSE,
                scale.value = 1, v4.4compat = FALSE)
{
  message("Beginning Cgee S-function, @(#) geeformula.q 4.13 98/01/27")
  call <- match.call()
  m <- match.call(expand.dots = FALSE)
  m$R <- m$b <- m$tol <- m$maxiter <- m$link <- m$varfun <-
    m$corstr <- m$Mv <- m$silent <- m$contrasts <-
    m$family <- m$scale.fix <- m$scale.value <- m$v4.4compat <- NULL
  if(is.null(m$id)) m$id <- as.name("id")
  if(!is.null(m$na.action) && m$na.action != "na.omit") {
    warning("Only 'na.omit' is implemented for gee\ncontinuing with 'na.action=na.omit'")
    m$na.action <- as.name("na.omit")
  }
  m[[1]] <- as.name("model.frame")
  m <- eval(m, parent.frame())
  Terms <- attr(m, "terms")
  y <- as.matrix(model.extract(m, "response"))
  x <- model.matrix(Terms, m, contrasts)
  ## R addition
  Q <- qr(x)
  if(Q$rank < ncol(x)) stop("rank-deficient model matrix")
  ## end of R addition
  N <- rep(1, length(y))
  if (dim(y)[2]==2) {
    N <- as.vector(y %*% c(1,1))
    y <- y[,1]
  }
  else {
    if (dim(y)[2]>2)
      stop("Only binomial response matrices (2 columns)")
  }
  offset <- model.extract(m, offset)
  id <- model.extract(m, id)
  id_raw <- id
  print(class(id_raw))
  if(is.null(id)) {
    stop("Id variable not found")
  }
  nobs <- nrow(x)
  p <- ncol(x)
  nclust <- sum(diff(as.integer(as.factor(id))) != 0) + 1
  xnames <- dimnames(x)[[2]]
  if(is.null(xnames)) {
    xnames <- paste("x", 1:p, sep = "")
    dimnames(x) <- list(NULL, xnames)
  }
  if (is.character(family)) family <- get(family)
  if (is.function(family)) family <- family()
  if (!is.null(b))
  {
    beta <- matrix(as.double(b), ncol = 1)
    if(nrow(beta) != p) {
      stop("Dim beta != ncol(x)")
    }
    message("user\'s initial regression estimate")
    print(beta)
  }
  else {
    message("running glm to get initial regression estimate")
    ### <tsl>	beta <- as.numeric(glm(m, family = family)$coef)
    mm <- match.call(expand.dots = FALSE)
    mm$R <- mm$b <- mm$tol <- mm$maxiter <- mm$link <- mm$varfun <-mm$corstr <- mm$Mv <- mm$silent <- mm$contrasts <-mm$scale.fix <- mm$scale.value <- mm$id<-NULL
    mm[[1]]<-as.name("glm")
    beta <- eval(mm, parent.frame())$coef
    ### </tsl>
    print(beta)
    beta <- as.numeric(beta)
  }
  if(length(id) != length(y))  stop("Id and y not same length")
  maxclsz <- as.integer(max(unlist(lapply(split(id, id), "length"))))
  maxiter <- as.integer(maxiter)
  silent <- as.integer(silent)
  if(length(offset) <= 1)  offset <- rep(0, length(y))
  if(length(offset) != length(y)) stop("offset and y not same length")
  offset <- as.double(offset)
  if(!is.null(R)) {
    Rr <- nrow(R)
    if(Rr != ncol(R)) stop("R is not square!")
    if(Rr < maxclsz) stop("R not big enough to accommodate some clusters.")
  }
  else {
    R <- matrix(as.double(rep(0, maxclsz * maxclsz)), nrow = maxclsz)
  }
  links <- c("identity", "log", "logit", "inverse", "probit",
             "cloglog")
  fams <- c("gaussian", "poisson", "binomial", "Gamma", "quasi")
  varfuns <- c("constant", "mu", "mu(1-mu)", "mu^2")
  corstrs <- c("independence", "fixed", "stat_M_dep", "non_stat_M_dep",
               "exchangeable", "AR-M", "unstructured")
  linkv <- as.integer(match(c(family$link), links, -1))
  famv <- match(family$family, fams, -1)
  if(famv < 1) stop("unknown family")
  if(famv <= 4) varfunv <- famv
  else varfunv <- match(family$varfun, varfuns, -1)
  varfunv <- as.integer(varfunv)
  corstrv <- as.integer(match(corstr, corstrs, -1))
  if(linkv < 1)
    stop("unknown link.")
  if(varfunv < 1)
    stop("unknown varfun.")
  if(corstrv < 1)
    stop("unknown corstr.")
  naivvar <- matrix(rep(0, p * p), nrow = p)
  robvar <- matrix(rep(0, p * p), nrow = p)
  score <- matrix(rep(0, nclust * p), ncol = p)
  hess <- matrix(rep(0, p * p), nrow = p)
  phi <- as.double(scale.value)
  scale.fix <- as.integer(scale.fix)
  errstate <- as.integer(1)
  tol <- as.double(tol)
  Mv <- as.integer(Mv)
  maxcl <- as.integer(0)
  if(!(is.double(x)))
    x <- as.double(x)
  if(!(is.double(y)))
    y <- as.double(y)
  if(!(is.double(id)))
    id <- as.double(id)
  if(!(is.double(N)))
    N <- as.double(N)
  modvec <- as.integer(c(linkv, varfunv, corstrv))
  if (v4.4compat) compatflag <- 1
  else compatflag <- 0
  z <- .C('Cgee',
          x,
          y,
          id,
          N,
          offset,
          nobs,
          p,
          modvec,
          Mv,
          estb = beta,
          nv = naivvar,
          rv = robvar,
          score = score,
          hess = hess,
          sc = phi,
          wcor = R,
          tol,
          mc = maxcl,
          iter = maxiter,
          silent,
          err = errstate,
          scale.fix,
          as.integer(compatflag))
  if(z$err != 0)
    warning("Cgee had an error (code= ", z$err, ").  Results suspect.")
  if(min(eigen(z$wcor)$values) < 0) {
    warning("Working correlation estimate not positive definite")
    z$err <- z$err + 1000
  }
  fit <- list()
  attr(fit, "class") <- c("gee", "glm")
  fit$title <- "GEE:  GENERALIZED LINEAR MODELS FOR DEPENDENT DATA"
  fit$version <- "gee S-function, version 4.13 modified 98/01/27 (1998)"
  links <- c("Identity", "Logarithm", "Logit", "Reciprocal", "Probit",
             "Cloglog")
  varfuns <- c("Gaussian", "Poisson", "Binomial", "Gamma")
  corstrs <- c("Independent", "Fixed", "Stationary M-dependent",
               "Non-Stationary M-dependent", "Exchangeable", "AR-M",
               "Unstructured")
  fit$model <- list()
  fit$model$link <- links[linkv]
  fit$model$varfun <- varfuns[varfunv]
  fit$model$corstr <- corstrs[corstrv]
  if(!is.na(match(c(corstrv), c(3, 4, 6))))
    fit$model$M <- Mv
  fit$call <- call
  fit$terms <- Terms
  fit$formula <- as.vector(attr(Terms, "formula"))
  fit$contrasts <- attr(x, "contrasts")
  fit$nobs <- nobs
  fit$iterations <- z$iter
  fit$coefficients <- as.vector(z$estb)
  fit$nas <- is.na(fit$coefficients)
  names(fit$coefficients) <- xnames
  eta <- as.vector(x %*% fit$coefficients)
  fit$linear.predictors <- eta
  ##Rchange
  mu <- as.vector(family$linkinv(eta))
  ##
  fit$fitted.values <- mu
  fit$residuals <- y - mu
  fit$family <- family
  fit$y <- as.vector(y)
  fit$id <- as.vector(id)
  fit$score_id <- rle(as.character(id_raw))$values
  fit$max.id <- z$mc
  z$wcor <- matrix(z$wcor, ncol = maxclsz)
  fit$working.correlation <- z$wcor
  fit$scale <- z$sc
  fit$score <- z$score
  stopifnot(length(fit$score_id) == nrow(fit$score))
  fit$hess <- z$hess
  fit$robust.variance <- z$rv # solve(hess)%*% (t(score)%*% score) %*% solve(hess)
  fit$naive.variance <- z$nv
  fit$xnames <- xnames
  fit$error <- z$err
  dimnames(fit$robust.variance) <- list(xnames, xnames)
  dimnames(fit$naive.variance) <- list(xnames, xnames)
  fit
}

#' Calculate Variance-Covariance Matrix for a Fitted Model Object
#' @description
#' Returns the variance-covariance matrix of the main parameters of fitted model 
#' objects. The "main" parameters of models correspond to those returned by `coef`. 
#' 
#' @param object an object returned by `jointCovariance()`.
#' @param model_index `NULL` if displaying covariance matrix of all fitted models;
#' otherwise, an integer indicating the fitted model.
#' @param ... for debugging only
#'
#' @return a matrix of covariance of all estimates
#' @export
#'
vcov.jointCovariance <- function(object, model_index = NULL, ...){
  
  ret <- object$mcov
  
  if(!is.null(model_index)){
    id_map <- object$id_map
    if(model_index > length(id_map)){
      stop('Invalid model index.')
    }
    
    ret <- ret[id_map[[model_index]], id_map[[model_index]], drop = FALSE]
    rownames(ret) <- names(id_map[[model_index]])
    colnames(ret) <- names(id_map[[model_index]])
  }
  
  ret
  
}

#' Rectal Indomethacin for Prevention of Post-ERCP Pancreatitis
#'
#' @description
#' Patient-level data from a multi-center, randomized, double-blind,
#' placebo-controlled 2-arm trial (n = 602) of rectal indomethacin (100 mg)
#' versus placebo to prevent post-ERCP pancreatitis in high-risk patients,
#' as reported by Elmunzer, Higgins, et al. (2012) in the
#' \emph{New England Journal of Medicine}.
#'
#' This dataset was originally collected, cleaned, reformatted, and released
#' for public teaching and research use by Dr. Peter D. R. Higgins in the
#' \pkg{medicaldata} R package as \code{indo_rct}. The version shipped here is
#' redistributed in support of the worked examples in this package. The
#' variable definitions below follow \pkg{medicaldata}; note that a few
#' columns are stored as numeric (rather than factor) in this copy. Users who
#' need the authoritative copy or accompanying documentation should consult
#' \pkg{medicaldata}.
#'
#' @name indo
#' @docType data
#' @format A data frame with 602 observations on the following 33 variables:
#' \describe{
#'   \item{\code{id}}{Subject identifier (numeric); leading digit indicates
#'     center. Range 1001--4003.}
#'   \item{\code{site}}{Study site (factor, 4 levels):
#'     \code{1_UM} = University of Michigan,
#'     \code{2_IU} = Indiana University,
#'     \code{3_UK} = University of Kentucky,
#'     \code{4_Case} = Case Western Reserve University.}
#'   \item{\code{age}}{Age in years (numeric), range 19--90.}
#'   \item{\code{risk}}{Risk score for post-ERCP pancreatitis (numeric),
#'     range 1--5.5.}
#'   \item{\code{gender}}{Sex (factor): \code{1_female}, \code{2_male}.}
#'   \item{\code{outcome}}{Primary outcome: post-ERCP pancreatitis
#'     (numeric, 1 = yes, 0 = no).}
#'   \item{\code{sod}}{Sphincter of Oddi dysfunction present (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{pep}}{History of prior post-ERCP pancreatitis (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{recpanc}}{History of recurrent pancreatitis (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{psphinc}}{Pancreatic sphincterotomy performed (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{precut}}{Sphincter pre-cut needed to enter papilla (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{difcan}}{Cannulation of papilla was difficult (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{pneudil}}{Pneumatic dilation of papilla performed (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{amp}}{Ampullectomy performed (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{paninj}}{Contrast injected into pancreas (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{acinar}}{Pancreas appeared to have acinarization on imaging
#'     (factor): \code{0_no}, \code{1_yes}.}
#'   \item{\code{brush}}{Brushings taken from pancreatic duct (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{asa81}}{Aspirin used at 81 mg per day (factor with 3 levels):
#'     \code{0_no}, \code{1_yes}, and a third level retained from the
#'     source coding.}
#'   \item{\code{asa325}}{Aspirin used at 325 mg per day (factor with 3
#'     levels): \code{0_no}, \code{1_yes}, and a third level retained from
#'     the source coding.}
#'   \item{\code{asa}}{Aspirin used at any dose (factor with 3 levels):
#'     \code{0_no}, \code{1_yes}, and a third level retained from the
#'     source coding.}
#'   \item{\code{prophystent}}{Pancreatic duct stent placed per endoscopist
#'     judgment (factor): \code{0_no}, \code{1_yes}.}
#'   \item{\code{therastent}}{Pancreatic duct stent placed to treat narrowing
#'     (factor): \code{0_no}, \code{1_yes}.}
#'   \item{\code{pdstent}}{Pancreatic duct stent placed for any reason
#'     (factor): \code{0_no}, \code{1_yes}.}
#'   \item{\code{sodsom}}{Sphincter of Oddi manometry performed (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{bsphinc}}{Biliary sphincterotomy performed (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{bstent}}{Biliary stent placed to relieve obstruction
#'     (factor): \code{0_no}, \code{1_yes}.}
#'   \item{\code{chole}}{Choledocholithiasis present (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{pbmal}}{Biliary duct or pancreatic malignancy found (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{train}}{Trainee participated in ERCP (factor):
#'     \code{0_no}, \code{1_yes}.}
#'   \item{\code{status}}{Patient status (factor):
#'     \code{0_inpatient}, \code{1_outpatient}.}
#'   \item{\code{type}}{Sphincter of Oddi dysfunction type (factor):
#'     \code{0_no SOD}, \code{1_type 1}, \code{2_type 2}, \code{3_type 3}.}
#'   \item{\code{rx}}{Treatment assignment (numeric, 1 = indomethacin,
#'     0 = placebo).}
#'   \item{\code{bleed}}{Reportable gastrointestinal bleeding (numeric,
#'     coded 1 = no, 2 = yes; \code{NA} when not assessed).}
#' }
#'
#' @source
#' Higgins, P. D. R. \pkg{medicaldata}: Data Package for Medical Datasets.
#' R package, dataset \code{indo_rct}.
#' \url{https://CRAN.R-project.org/package=medicaldata}
#'
#' @references
#' Elmunzer BJ, Higgins PDR, Saini SD, et al. A randomized trial of rectal
#' indomethacin to prevent post-ERCP pancreatitis.
#' \emph{New England Journal of Medicine} 2012; 366(15):1414--1422.
#' \doi{10.1056/NEJMoa1111103}
#'
#' @examples
#' data(indo)
#'
NULL

asymptoticJointCovariance <- function(..., data){
  
  specs <- list(...)
  if(length(specs) == 0){
    stop("No model specs provided in ...")
  }
  
  if(!is.list(data)){
    stop("data must be a list of data.frames.")
  }
  
  adapters <- lapply(specs, make_adapter, data_list = data)
  adapters <- lapply(adapters, function(ad){ ad$fit_model(); ad })
  
  n_models <- length(adapters)
  bet <- NULL
  df <- NULL
  vars <- vector('list', n_models)
  for(i in 1:n_models){
    engine <- adapters[[i]]$get_engine()
    if(engine %in% c('km', 'quantile')){
      stop('Model ', i, ' uses engine "', engine, '_" which has no closed-form score. ',
           'Re-fit with nboot > 0 to use the bootstrap path.')
    }
    fit <- adapters[[i]]$get_coef()
    names(fit) <- paste0('model_', i, '_', names(fit))
    bet <- c(bet, fit)
    df <- c(df, adapters[[i]]$get_dim())
    vars[[i]] <- adapters[[i]]$get_coef_name()
  }
  
  
  n_par <- length(bet)
  mcov <- matrix(0, n_par, n_par)
  n_shared_sample_sizes <- matrix(0, n_models, n_models)
  id_map <- NULL
  
  for(i in 1:n_models){
    idx1 <- adapters[[i]]$get_data_index()
    n_shared_sample_sizes[i, i] <- adapters[[i]]$get_n()
    score1 <- adapters[[i]]$get_score()
    bread1 <- adapters[[i]]$get_bread()
    
    id1 <- IDMapping(df, i, vars[[i]])
    id_map[[i]] <- id1
    
    for(j in i:n_models){
      idx2 <- adapters[[j]]$get_data_index()
      n_shared_sample_sizes[i, j] <- intersect(adapters[[i]]$get_sample_id(), adapters[[j]]$get_sample_id()) %>% length()
      n_shared_sample_sizes[j, i] <- n_shared_sample_sizes[i, j]
      
      score2 <- adapters[[j]]$get_score()
      bread2 <- adapters[[j]]$get_bread()
      
      id2 <- IDMapping(df, j, vars[[j]])
      info <- FisherInformation(score1, score2)
      mcov[id1, id2] <-
        bread1 %*%
        info %*%
        t(bread2) *
        (attr(info, 'n') / nrow(score1) / nrow(score2))
      mcov[id2, id1] <- t(mcov[id1, id2])
      
      rm(score2)
    }
    rm(score1)
    
  }
  
  bet <- unname(bet)
  # rownames(mcov) <- names(bet)
  # colnames(mcov) <- names(bet)
  
  fit <- 
    list(
      coefficients = bet,
      mcov = mcov,
      id_map = id_map,
      n_shared_sample_sizes = n_shared_sample_sizes
    )
  
  invisible(fit)
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/actg.R
\docType{data}
\name{actg}
\alias{actg}
\title{ACTG 320 Clinical Trial Dataset}
\format{
A data frame
\describe{
\item{id}{Identification Code}
\item{time}{Time to AIDS diagnosis or death (days).}
\item{censor}{Event indicator. 1 = AIDS defining diagnosis, 0 = Otherwise.}
\item{time_d}{Time to death (days)}
\item{censor_d}{Event indicator for death (only). 1 = Death, 0 = Otherwise.}
\item{tx}{Treatment indicator. 1 = Treatment includes IDV, 0 = Control group.}
\item{txgrp}{Treatment group indicator. 1 = ZDV + 3TC. 2 = ZDV + 3TC + IDV. 3 = d4T + 3TC. 4 = d4T + 3TC + IDV.}
\item{strat2}{CD4 stratum at screening. 0 = CD4 <= 50. 1 = CD4 > 50.}
\item{sex}{0 = Male. 1 = Female.}
\item{raceth}{Race/Ethnicity. 1 = White Non-Hispanic. 2 = Black Non-Hispanic. 3 = Hispanic. 4 = Asian, Pacific Islander. 5 = American Indian, Alaskan Native. 6 = Other/unknown.}
\item{ivdrug}{IV drug use history. 1 = Never. 2 = Currently. 3 = Previously.}
\item{hemophil}{Hemophiliac. 1 = Yes. 0 = No.}
\item{karnof}{Karnofsky Performance Scale. 100 = Normal; no complaint no evidence of disease. 90 = Normal activity possible; minor signs/symptoms of disease. 80 = Normal activity with effort; some signs/symptoms of disease. 70 = Cares for self; normal activity/active work not possible.}
\item{cd4}{Baseline CD4 count (Cells/Milliliter).}
\item{priorzdv}{Months of prior ZDV use (months).}
\item{age}{Age at Enrollment (years).}
}
}
\source{
ftp://ftp.wiley.com/public/sci_tech_med/survival
}
\description{
actg dataset from Hosmer et al.
}
\examples{
data(actg)

}
\references{
Hosmer, D.W. and Lemeshow, S. and May, S. (2008) Applied Survival Analysis: Regression Modeling of Time to Event Data: Second Edition, John Wiley and Sons Inc., New York, NY
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.jointCovariance.R
\name{summary.jointCovariance}
\alias{summary.jointCovariance}
\title{Object Summaries}
\usage{
\method{summary}{jointCovariance}(object, model_index = NULL, ...)
}
\arguments{
\item{object}{an object returned by \code{jointCovariance()}.}

\item{model_index}{\code{NULL} if displaying summary of all fitted models; otherwise,
an integer indicating the fitted model.}

\item{...}{for debugging only}
}
\value{
a list
}
\description{
\code{summary} method for class \code{jointCovariance}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coxph_.R
\name{coxph_}
\alias{coxph_}
\title{Creating Objects of Proportional Hazards Regression Model}
\usage{
coxph_(formula, data_index = 1)
}
\arguments{
\item{formula}{see \code{formula} in \code{survival::coxph}.}

\item{data_index}{integer. Index of the data frame in the \code{data} argument of
\code{jointCovariance} to be used when fitting a proportional hazards model.}
}
\description{
\code{coxph_} is a wrapper function of \code{survival::coxph} to create an object to be
passed into \code{jointCovariance}, the main function of this package through its
argument \code{...}. The object defines how a proportional hazard model would be
fitted.
}
\details{
Not all arguments of \code{survival::coxph} are supported in \code{coxph_} due to the
complexity in handling environment and scope, which is particularly difficult
for arguments like \code{weights}, \code{subset}, etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mmrm_.R
\name{mmrm_}
\alias{mmrm_}
\title{Creating Objects of Mixed Models for Repeated Measures}
\usage{
mmrm_(
  formula,
  covariance = NULL,
  reml = TRUE,
  control = mmrm::mmrm_control(...),
  ...,
  data_index = 1
)
}
\arguments{
\item{formula}{see \code{formula} in \code{mmrm::mmrm}.}

\item{covariance}{see \code{covariance} in \code{mmrm::mmrm}.}

\item{reml}{see \code{reml} in \code{mmrm::mmrm}.}

\item{control}{see \code{control} in \code{mmrm::mmrm}.}

\item{...}{see \code{...} in \code{mmrm::mmrm}.}

\item{data_index}{integer. Index of the data frame in the \code{data} argument of
\code{jointCovariance} to be used when fitting a GEE model.}
}
\description{
\code{mmrm_} is a wrapper function of \code{mmrm::mmrm} to create an object to be passed
into \code{jointCovariance}, the main function of this package through its
argument \code{...}. The object defines how a MMRM model would be fitted.
}
\details{
The argument \code{weights} of \code{mmrm::mmrm} is supported in \code{mmrm_} due to the
complexity in handling environment and scope.

Please always refer to help document of \code{mmrm::mmrm} before using \code{mmrm_}.
For example, time variable and observation ID must be factor variables in
some cases, otherwise error may be prompted. Users can call \code{mmrm::mmrm}
using the same arguments being passed to \code{mmrm_} to check validity.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vcov.jointCovariance.R
\name{vcov.jointCovariance}
\alias{vcov.jointCovariance}
\title{Calculate Variance-Covariance Matrix for a Fitted Model Object}
\usage{
\method{vcov}{jointCovariance}(object, model_index = NULL, ...)
}
\arguments{
\item{object}{an object returned by \code{jointCovariance()}.}

\item{model_index}{\code{NULL} if displaying covariance matrix of all fitted models;
otherwise, an integer indicating the fitted model.}

\item{...}{for debugging only}
}
\value{
a matrix of covariance of all estimates
}
\description{
Returns the variance-covariance matrix of the main parameters of fitted model
objects. The "main" parameters of models correspond to those returned by \code{coef}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/indo.R
\docType{data}
\name{indo}
\alias{indo}
\title{Rectal Indomethacin for Prevention of Post-ERCP Pancreatitis}
\format{
A data frame with 602 observations on the following 33 variables:
\describe{
\item{\code{id}}{Subject identifier (numeric); leading digit indicates
center. Range 1001--4003.}
\item{\code{site}}{Study site (factor, 4 levels):
\code{1_UM} = University of Michigan,
\code{2_IU} = Indiana University,
\code{3_UK} = University of Kentucky,
\code{4_Case} = Case Western Reserve University.}
\item{\code{age}}{Age in years (numeric), range 19--90.}
\item{\code{risk}}{Risk score for post-ERCP pancreatitis (numeric),
range 1--5.5.}
\item{\code{gender}}{Sex (factor): \code{1_female}, \code{2_male}.}
\item{\code{outcome}}{Primary outcome: post-ERCP pancreatitis
(numeric, 1 = yes, 0 = no).}
\item{\code{sod}}{Sphincter of Oddi dysfunction present (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{pep}}{History of prior post-ERCP pancreatitis (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{recpanc}}{History of recurrent pancreatitis (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{psphinc}}{Pancreatic sphincterotomy performed (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{precut}}{Sphincter pre-cut needed to enter papilla (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{difcan}}{Cannulation of papilla was difficult (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{pneudil}}{Pneumatic dilation of papilla performed (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{amp}}{Ampullectomy performed (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{paninj}}{Contrast injected into pancreas (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{acinar}}{Pancreas appeared to have acinarization on imaging
(factor): \code{0_no}, \code{1_yes}.}
\item{\code{brush}}{Brushings taken from pancreatic duct (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{asa81}}{Aspirin used at 81 mg per day (factor with 3 levels):
\code{0_no}, \code{1_yes}, and a third level retained from the
source coding.}
\item{\code{asa325}}{Aspirin used at 325 mg per day (factor with 3
levels): \code{0_no}, \code{1_yes}, and a third level retained from
the source coding.}
\item{\code{asa}}{Aspirin used at any dose (factor with 3 levels):
\code{0_no}, \code{1_yes}, and a third level retained from the
source coding.}
\item{\code{prophystent}}{Pancreatic duct stent placed per endoscopist
judgment (factor): \code{0_no}, \code{1_yes}.}
\item{\code{therastent}}{Pancreatic duct stent placed to treat narrowing
(factor): \code{0_no}, \code{1_yes}.}
\item{\code{pdstent}}{Pancreatic duct stent placed for any reason
(factor): \code{0_no}, \code{1_yes}.}
\item{\code{sodsom}}{Sphincter of Oddi manometry performed (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{bsphinc}}{Biliary sphincterotomy performed (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{bstent}}{Biliary stent placed to relieve obstruction
(factor): \code{0_no}, \code{1_yes}.}
\item{\code{chole}}{Choledocholithiasis present (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{pbmal}}{Biliary duct or pancreatic malignancy found (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{train}}{Trainee participated in ERCP (factor):
\code{0_no}, \code{1_yes}.}
\item{\code{status}}{Patient status (factor):
\code{0_inpatient}, \code{1_outpatient}.}
\item{\code{type}}{Sphincter of Oddi dysfunction type (factor):
\code{0_no SOD}, \code{1_type 1}, \code{2_type 2}, \code{3_type 3}.}
\item{\code{rx}}{Treatment assignment (numeric, 1 = indomethacin,
0 = placebo).}
\item{\code{bleed}}{Reportable gastrointestinal bleeding (numeric,
coded 1 = no, 2 = yes; \code{NA} when not assessed).}
}
}
\source{
Higgins, P. D. R. \pkg{medicaldata}: Data Package for Medical Datasets.
R package, dataset \code{indo_rct}.
\url{https://CRAN.R-project.org/package=medicaldata}
}
\description{
Patient-level data from a multi-center, randomized, double-blind,
placebo-controlled 2-arm trial (n = 602) of rectal indomethacin (100 mg)
versus placebo to prevent post-ERCP pancreatitis in high-risk patients,
as reported by Elmunzer, Higgins, et al. (2012) in the
\emph{New England Journal of Medicine}.

This dataset was originally collected, cleaned, reformatted, and released
for public teaching and research use by Dr. Peter D. R. Higgins in the
\pkg{medicaldata} R package as \code{indo_rct}. The version shipped here is
redistributed in support of the worked examples in this package. The
variable definitions below follow \pkg{medicaldata}; note that a few
columns are stored as numeric (rather than factor) in this copy. Users who
need the authoritative copy or accompanying documentation should consult
\pkg{medicaldata}.
}
\examples{
data(indo)

}
\references{
Elmunzer BJ, Higgins PDR, Saini SD, et al. A randomized trial of rectal
indomethacin to prevent post-ERCP pancreatitis.
\emph{New England Journal of Medicine} 2012; 366(15):1414--1422.
\doi{10.1056/NEJMoa1111103}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pated.R
\name{plot.pated}
\alias{plot.pated}
\title{Plot PATED Analysis Results}
\usage{
\method{plot}{pated}(x, ...)
}
\arguments{
\item{x}{an object returned from \code{pated()}.}

\item{...}{currently not supported.}
}
\value{
\code{NULL}
}
\description{
Plot PATED Analysis Results
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glm_.R
\name{glm_}
\alias{glm_}
\title{Creating Objects of Generalized Linear Models}
\usage{
glm_(formula, family, data_index = 1)
}
\arguments{
\item{formula}{see \code{formula} in \code{stats::glm}.}

\item{family}{currently supports \code{"gaussian"} or \code{"binomial"}. Other families
are under testing.}

\item{data_index}{integer. Index of the data frame in the \code{data} argument of
\code{jointCovariance} to be used when fitting a generalized linear model.}
}
\description{
\code{glm_} is a wrapper function of \code{stats::glm} to create an object to be passed
into \code{jointCovariance}, the main function of this package through its
argument \code{...}. The object defines how a GLM model would be fitted.
}
\details{
Not all arguments of \code{stats::glm} are supported in \code{glm_} due to the
complexity in handling environment and scope, which is particularly difficult
for arguments like \code{weights}, \code{subset}, etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jointCovariance.R
\name{jointCovariance}
\alias{jointCovariance}
\title{Fitting Regression Models for Multiple Outcomes and Returning the Matrix of
Covariance}
\usage{
jointCovariance(..., data, nboot = 0, compute_cov = TRUE, seed = NULL)
}
\arguments{
\item{...}{objects returned by \code{glm_()}, \code{coxph_()}, \code{logrank_()}, \code{gee_()}
and \code{gmm_()}.}

\item{data}{a data frame if all models are fitted on the same dataset;
otherwise a list of data frames for fitting models in \code{...}. Note that a
dataset can be used to fit multiple models, thus, \code{length(data)} is unnecessary
to be equal to the number of models in \code{...}. The row names in a data frame
are not treated as subject IDs. Instead, all data frame should consist of
a column \code{pid} as subject IDs. For any two records in different
data frames that correspond to the same subject, their values in \code{pid} should
be consistent. For data frames to be used by GEE, \code{pid} defines clusters.
See \code{id} and the Details section of \code{gee:gee}.}

\item{nboot}{non-zero integer if bootstrap is adopted. By default 0.}

\item{compute_cov}{logic. If \code{TRUE} and \code{nboot > 0},
empirical covariance matrix is computed using bootstrap estimate and
returned. Bootstrap estimate will be abandoned. If
\code{FALSE}, bootstrap estimate will be returned and no empirical covariance
matrix is computed.}

\item{seed}{random seed when generate bootstrap data.}
}
\value{
It returns an object of class "jointCovariance", which is a list
containing the following components:
\tabular{ll}{
\code{coefficients} \tab an unnamed vector of coefficients of all fitted models.
Use \code{id_map} for variable mapping. \cr
\tab \cr
\code{mcov} \tab a unnamed matrix of covariance of \code{coefficients}. Use \code{id_map}
for variable mapping. \cr
\tab \cr
\code{id_map} \tab a list mapping the elements in \code{coefficients} and \code{mcov} to
variable names. \cr
\tab \cr
\code{n_shared_sample_sizes} \tab a matrix of shared sample sizes between datasets
being used to fit the models. \cr
\tab \cr
\code{call} \tab the matched call.\cr
}
}
\description{
\code{jointCovariance} can fit different types of models for multiple outcomes
simultaneously and return model parameters and variance-covariance matrix
for further analysis.
}
\examples{
## More examples can be found in the vignettes.
library(survival)
library(mvtnorm)
library(tidyr)
genData <- function(seed = NULL){
  
  set.seed(seed)
  n <- 300
  sigma <- matrix(.7, 4, 4)
  diag(sigma) <- 1
  v <- rmvnorm(n, sigma = sigma)
  x1 <- v[, 1]
  x2 <- v[, 2]
  z1 <- (v[, 3] > 0) + 0
  z2 <- v[, 4]
  
  trt <- rbinom(n, 1, .5)
  
  bet <- c(-.3,.3)
  y <- -log(runif(n))/
    exp(-.3 * x1 + .3 * x2 + z1 * .5 - z2 * .3 + .1 * trt + rnorm(n))
  
  z1[sample.int(n, 50)] <- NA
  z2[sample.int(n, 50)] <- NA
  x1[sample.int(n, 50)] <- NA
  x2[sample.int(n, 50)] <- NA
  death <- ifelse(y > 2, 0, 1)
  y[y > 2] <- 2
  
  pid <- paste0('pid-', 1:n)
  ret <- data.frame(
    y = y, trt = trt, 
    z1 = z1, z2 = z2, 
    x1 = x1, x2 = x2, 
    death, pid)
  ret
}

dat1 <- genData()

## create a dataset with repeated measurements x
dat2 <- dat1 \%>\% pivot_longer(c(x1, x2), names_to='visit', values_to='x') \%>\% 
  dplyr::select(x, trt, visit, pid) \%>\% as.data.frame()

dat2$visit <- as.factor(dat2$visit)
dat2$pid <- as.factor(dat2$pid)

fit <- jointCovariance(
  coxph_(Surv(time = y, event = death) ~ trt, data_index = 1),
  logrank_(Surv(time = y, event=death) ~ trt, data_index = 1),
  glm_(z1 ~ trt, family = 'binomial', data_index = 1), 
  glm_(z2 ~ trt, family = 'gaussian', data_index = 1), 
  mmrm_(x ~ trt + us(visit | pid), reml = TRUE, data_index = 2), 
  gee_(x ~ trt, family = 'gaussian', corstr = 'independence', data_index = 2), 
  data = list(dat1, dat2))

fit

\donttest{
bfit <-jointCovariance(
  coxph_(Surv(time=y, event=death) ~ trt, data_index = 1),
  logrank_(Surv(time=y, event=death) ~ trt, data_index = 1),
  glm_(z1 ~ trt, family = 'binomial', data_index = 1),
  glm_(z2 ~ trt, family = 'gaussian', data_index = 1),
  mmrm_(x ~ trt + us(visit | pid), reml = TRUE, data_index = 2),
  gee_(x ~ trt, family = 'gaussian', corstr = 'independence', data_index = 2),
  data = list(dat1, dat2), nboot = 10)

summary(bfit)

## km_() and quantile_() require nboot > 0 because they have no
## closed-form score. compute_cov is forced to FALSE for km_().
## When all models share one dataset, `data_index` and `list(...)`
## can be omitted.
kfit <- jointCovariance(
  km_(Surv(time = y, event = death) ~ trt, conf_type = 'log',
      times = c(0.5, 1, 1.5)),
  glm_(z1 ~ trt, family = 'binomial'),
  data = dat1, nboot = 30, seed = 1)

qfit <- jointCovariance(
  quantile_(y ~ trt, probs = c(0.25, 0.5, 0.75)),
  glm_(z2 ~ trt, family = 'gaussian'),
  data = dat1, nboot = 30, seed = 1)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pated.R
\name{pated}
\alias{pated}
\title{Prognostic Variables Assisted Treatment Effect Detection}
\usage{
pated(
  ...,
  data,
  nboot = 0,
  compute_cov = TRUE,
  seed = NULL,
  transform = "identity"
)
}
\arguments{
\item{...}{model specifications built by \code{glm_()}, \code{coxph_()}, \code{logrank_()},
\code{gee_()}, \code{mmrm_()}, \code{km_()}, or \code{quantile_()}. The first specification is
the primary outcome whose treatment effect is being tested; the rest are
prognostic covariates used to tighten the SE.}

\item{data}{either a single data frame (when all models are fitted on
the same dataset) or a list of data frames (one entry per \code{data_index}).
Each data frame must have a \code{pid} column carrying subject identifiers;
records with the same \code{pid} across different data frames refer to the
same subject.}

\item{nboot}{non-zero integer if bootstrap is adopted. By default 0.}

\item{compute_cov}{logic. If \code{TRUE}, empirical covariance matrix is computed
using bootstrap estimate and returned. Bootstrap estimate will be abandoned. If
\code{FALSE}, bootstrap estimate will be returned and no empirical covariance
matrix is computed.}

\item{seed}{random seed when generate bootstrap data.}

\item{transform}{character. Now only supports \code{"identity"}.}
}
\value{
a data frame of testing results.
}
\description{
\code{pated} is a wrapper function of \code{jointCovariance} for testing treatment effect
in randomized clinical trials. It assumes that prognostic variables are fully
randomized. This assumption can help enhancing statistical power of conventional
approaches in detecting the treatment effect. Specifically, the sensitivity
of the conventional models specified in \code{...} are improved by \code{pated}.
}
\examples{
## More examples can be found in the vignettes.
library(survival)
library(mvtnorm)

genData <- function(seed = NULL){
  set.seed(seed)
  n <- 300
  sigma <- matrix(c(1, .6, .6, 1), 2)
  x <- rmvnorm(n, sigma = sigma)
  z1 <- rbinom(n, 1, .6)
  z2 <- rnorm(n)
  trt <- rbinom(n, 1, .5)

  bet <- c(-.2, .2)
  y <- -.5 + x \%*\% bet + z1 * .3 - z2 * .1 + .1 * trt - .1 * rnorm(n)
  death <- rbinom(n, 1, .8)
  data.frame(
    y = as.numeric(y), trt = trt,
    z1 = z1, z2 = z2,
    x1 = x[, 1], x2 = x[, 2],
    death, pid = paste0('s-', seq_len(n))
  )
}

dat <- genData(seed = 31415926)

## `data_index` defaults to 1 in every spec constructor and a single
## data.frame is auto-wrapped into a list, so neither needs spelling out
## when all models are fitted on the same dataset.
fit <-
  pated(
    coxph_(Surv(time = y, event = death) ~ trt),
    glm_(z1 ~ trt, family = 'binomial'),
    glm_(z2 ~ trt, family = 'gaussian'),
    glm_(x1 ~ trt, family = 'gaussian'),
    glm_(x2 ~ trt, family = 'gaussian'),
    data = dat
  )

fit

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/km_.R
\name{km_}
\alias{km_}
\title{Creating Objects of Kaplan-Meier Curve}
\usage{
km_(formula, times = NULL, conf_type, data_index = 1)
}
\arguments{
\item{formula}{a formula created by \code{survival::Surv()}.}

\item{times}{numeric vector of time. Survival probabilities at \code{times} are
computed (with g-transformation defined by \code{conf_type}).}

\item{conf_type}{character. Type of confidence interval. It must be one of
\code{"log"}, \code{"log-log"}, \code{"plain"}, \code{"logit"}, or \code{"arcsin"}.}

\item{data_index}{integer. Index of the data frame in the \code{data} argument of
\code{jointCovariance} to be used when fitting a generalized linear model.}
}
\description{
\code{km_} is a wrapper function creating an object of Kaplan-Meier curve to be
passed into \code{jointCovariance}, the main function of this package through its
argument \code{...}. The object defines how a Kaplan-Meier curve would be fitted.
}
\details{
Usually, g-transformation is applied to the survival probability \code{S(t)}
to obtain pointwise confidence interval of a Kaplan-Meier curve. This
can be achieved by specifying \code{conf_type}. For identity transformation,
use \code{conf_type = "plain"}.

This function can only be used with \code{jointCovariance} when the bootstrap
method is used to estimate variance-covariance matrix of multiple outcome
models.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gee_.R
\name{gee_}
\alias{gee_}
\title{Creating Objects of Generalized Estimation Equation Model}
\usage{
gee_(formula, family, corstr, R = NULL, b = NULL, Mv = 1, data_index = 1)
}
\arguments{
\item{formula}{see \code{formula} in \code{gee::gee}.}

\item{family}{see \code{family} in \code{gee::gee}.}

\item{corstr}{see \code{corstr} in \code{gee::gee}.}

\item{R}{see \code{R} in \code{gee::gee}.}

\item{b}{see \code{b} in \code{gee::gee}.}

\item{Mv}{see \code{Mv} in \code{gee::gee}.}

\item{data_index}{integer. Index of the data frame in the \code{data} argument of
\code{jointCovariance} to be used when fitting a GEE model.}
}
\description{
\code{gee_} is a wrapper function of \code{gee::gee} to create an object to be passed
into \code{jointCovariance}, the main function of this package through its
argument \code{...}. The object defines how a GEE model would be fitted.

This package does not import the package \code{gee}. Instead, codes of \code{gee} are
modified and integrated to compute score and information matrix. Thus, users
does not need to install the package \code{gee} to use this package.
}
\details{
Not all arguments of \code{stats::gee} are supported in \code{gee_} due to the
complexity in handling environment and scope, which is particularly difficult
for arguments like \code{subset}, etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.summary.jointCovariance.R
\name{print.summary.jointCovariance}
\alias{print.summary.jointCovariance}
\title{Title Summarize an Analysis of Multiple Outcomes.}
\usage{
\method{print}{summary.jointCovariance}(x, ...)
}
\arguments{
\item{x}{an object returned by \code{jointCovariance()}.}

\item{...}{for debugging only.}
}
\value{
an invisible object.
}
\description{
Summarize an analysis of multiple outcomes.
}
\examples{
## no example
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulateMoData.R
\name{simulateMoData}
\alias{simulateMoData}
\title{Generating Data for Simulation and Testing}
\usage{
simulateMoData(n = 500, hr = 0.8, seed = NULL)
}
\arguments{
\item{n}{an integer for total sample size of a randomized control trial of
two arms.}

\item{hr}{hazard ratio of treatment.}

\item{seed}{random seed. By default \code{NULL} for no seed being specified.}
}
\description{
\code{simulateMoData} generates data for simulation and testing purposes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quantile_.R
\name{quantile_}
\alias{quantile_}
\title{Creating Objects of Group Quantile Differences}
\usage{
quantile_(formula, probs = c(0.25, 0.5, 0.75), data_index = 1)
}
\arguments{
\item{formula}{a two-sided formula \code{y ~ arm} where the right-hand side is a
binary grouping variable.}

\item{probs}{numeric vector of probabilities in \verb{(0, 1)}. By default the
first quartile, median, and third quartile.}

\item{data_index}{integer. Index of the data frame in the \code{data} argument
of \code{jointCovariance} to be used.}
}
\description{
\code{quantile_} is a wrapper function creating an object that, for each
requested probability, computes the difference between the two arms'
sample quantiles of the outcome. The object is passed to \code{jointCovariance}
or \code{pated} through \code{...}. Because the empirical-quantile estimator has no
tractable closed-form score, this engine is available only through the
bootstrap path (\code{nboot > 0}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logrank_.R
\name{logrank_}
\alias{logrank_}
\title{Creating Objects of Logrank Test}
\usage{
logrank_(formula, ties = c("efron", "breslow", "exact"), data_index = 1)
}
\arguments{
\item{formula}{see \code{formula} in \code{survival::coxph}.}

\item{ties}{character string specifying the method for tie handling. One of
\code{"efron"} (default), \code{"breslow"}, or \code{"exact"}. Passed through to
\code{survival::coxph}.}

\item{data_index}{integer. Index of the data frame in the \code{data} argument of
\code{jointCovariance} to be used when computing testing statistic of logrank
test.}
}
\description{
\code{logrank_} is a wrapper function \code{survival::coxph} to create an object to be
passed into \code{jointCovariance}, the main function of this package through its
argument \code{...}. Logrank test is the score test under the proportional hazards
regression model. The object defines how a logrank test would be computed.
}
\details{
Not all arguments of \code{survival::coxph} are supported in \code{logrank_} due to the
complexity in handling environment and scope, which is particularly difficult
for arguments like \code{weights}, \code{subset}, etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coef.jointCovariance.R
\name{coef.jointCovariance}
\alias{coef.jointCovariance}
\title{Extract Model Coefficients}
\usage{
\method{coef}{jointCovariance}(object, model_index = NULL, ...)
}
\arguments{
\item{object}{an object returned by \code{jointCovariance()}.}

\item{model_index}{\code{NULL} if displaying coefficients of all fitted models;
otherwise, an integer indicating the fitted model.}

\item{...}{for debugging only}
}
\value{
a vector of coefficient estimates
}
\description{
\code{coef} is a generic function.
}
